//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   9 May 07  Andy Frank  Creation
//

using compiler
using fandoc

**
** SourceToHtmlGenerator generates an syntax color coded HTML
** file for a Type's source code.
**
class SourceToHtmlGenerator : ApiToHtmlGenerator
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  new make(DocCompiler compiler, Location loc, OutStream out, Type t, File srcFile)
    : super(compiler, loc, out, t)
  {
    this.srcFile = srcFile
  }

//////////////////////////////////////////////////////////////////////////
// ApiToHtmlGenerator
//////////////////////////////////////////////////////////////////////////

  **
  ** Generate the main content.
  **
  override Void content()
  {
    out.print("<div class='type'>\n")
    typeOverview
    out.print("</div>\n")
    out.print("<div class='src'>\n")
    out.print("<pre>")
    srcFileFacet := t->sourceFile
    file := t.pod.files["/src/$srcFileFacet".toUri]
    convert(file.in.readAllLines, out)
    out.print("</pre>\n")
    out.print("</div>\n")
  }

  **
  ** Generate the sidebar.
  **
  override Void sidebar()
  {
    out.print("<h2>More Info</h2>\n")
    out.print("<ul>\n")
    out.print("  <li><a href='${t.name}.html'>View Fandoc</a></li>\n")
    out.print("</ul>\n")
    slotsOverview(false)
  }

//////////////////////////////////////////////////////////////////////////
// FanToHtml
//////////////////////////////////////////////////////////////////////////

  **
  ** Convert the Fan source code to syntax highlighted HTML markup.
  **
  Void convert(Str[] lines, OutStream out)
  {
    max := lines.size.toStr.size
    inBlockComment := false


    lines.each |Str line, Int i|
    {
      num := i+1
      pad := max - num.toStr.size + 1

      out.print("<span class='a'")
      slot := t.slots.find |Slot s -> Bool| // TODO: this nested loop sucks
      {
        return s.parent == t && s->lineNumber == num
      }
      if (slot != null && slot.name != "instance\$init")
        out.print(" id='$slot.name'")
      out.print(">")
      out.print(Str.spaces(pad))
      out.print("${num} </span>")
      out.print(" ")
      inBlockComment = markup(line, out, inBlockComment)

      out.print("\n")
    }
  }

//////////////////////////////////////////////////////////////////////////
// Support Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Markup the given line with syntax highlighting.
  **
  private Bool markup(Str line, OutStream out, Bool inBlockComment := false)
  {
    inStr := false
    buf   := Buf.make

    if (inBlockComment) out.print("<span class='b'") // block comment

    for (i := 0; i<line.size; i++)
    {
      curr := line[i]
      peek := (i < line.size-1) ? line[i+1] : -1
      pre  := (i > 0) ? line[i-1] : -1

      // Currenty in a block comment, keep looping till we find the end
      if (inBlockComment)
      {
        safe(out, curr)
        if (curr == '*' && peek == '/')
        {
          inBlockComment = false
          safe(out, peek)
          out.print("</span>")
          i++
        }
        continue
      }

      // Currently in a Str or Uri, keep looping till we find the end
      if (inStr)
      {
        safe(out, curr)
        if ((curr == '"' && pre != '\\') || curr == '`')
        {
          inStr = false
          out.print("</span>")
        }
        continue
      }

      // We're starting a Str
      if (curr == '"')
      {
        inStr = true
        out.print("<span class='g'>").writeChar(curr)  // string
        continue
      }

      // We're starting a Uri
      if (curr == '`')
      {
        inStr = true
        out.print("<span class='i'>").writeChar(curr) // uri
        continue
      }

      // We're starting a Char
      if (curr == '\'')
      {
        out.print("<span class='h'>").writeChar(curr) // char
        safe(out, peek)
        if (peek == '\\')
        {
          safe(out, line[i+2])
          i++
        }
        i += 2
        out.writeChar('\'')
        out.print("</span>")
        continue
      }

      // Block comment
      if (curr == '/' && peek == '*')
      {
        out.print("<span class='b'>") // block comment
        inBlockComment = true
        out.writeChar(curr).writeChar(peek)
        i++
        continue
      }

      // Line comment
      if (curr == '/' && peek == '/')
      {
        out.print("<span class='c'>") // line comment
        safeStr(out, line[i..-1])
        out.print("</span>")
        return false
      }

      // Fandoc comment
      if (curr == '*' && peek == '*')
      {
        out.print("<span class='d'>") // fandoc comment
        safeStr(out, line[i..-1])
        out.print("</span>")
        return false
      }

      if (isAlphaNum(curr))
      {
        // Build up buffer to check if this is a keyword
        buf.writeChar(curr)
        continue
      }
      else if (!buf.empty)
      {
        // Word is over, write out checking for keywords
        doBuf(buf, out)

        // Reset buf for next time
        buf.clear
      }

      // Single character
      if (curr == '(' || curr == ')' ||
          curr == '{' || curr == '}' ||
          curr == '[' || curr == ']')
      {
        out.print("<span class='e'>").writeChar(curr) // bracket
        out.print("</span>")
      }
      else
      {
        safe(out, curr)
      }
    }

    // If buf ended the line, make sure we still check for a keyword
    if (!buf.empty) doBuf(buf, out)

    // If still in block comment, close tag for this line
    if (inBlockComment) out.print("</span>")
    return inBlockComment
  }

  **
  ** Convenience for s.each |Int ch| { safe(out, ch) }
  **
  private Void safeStr(OutStream out, Str s)
  {
    s.each |Int ch| { safe(out, ch) }
  }

  **
  ** Escape <, &, and > characters.
  **
  private Void safe(OutStream out, Int ch)
  {
    //if (ch == ' ') out.print("&nbsp;")
    if (ch == '<') out.print("&lt;")
    else if (ch == '>') out.print("&gt;")
    else if (ch == '&') out.print("&amp;")
    else out.writeChar(ch)
  }

  **
  ** Handle writing the buf, checking for keywords. This method
  ** assumes the buf is still in write mode.
  **
  private Void doBuf(Buf buf, OutStream out)
  {
    buf.flip
    s := buf.readAllStr

    if (isKeyword(s))
    {
      // Keyword
      out.print("<span class='f'>$s</span>") // keyword
    }
    else
    {
      // Indentifier
      out.print(s)
    }
  }

  **
  ** Return true if the buf contains only alphabetic characters,
  ** numerals or underscores.
  **
  private Bool isAlphaNum(Int ch)
  {
    return ch.isAlphaNum || ch === '_'
  }

  **
  ** Return true if this Str matches a keyword
  **
  private Bool isKeyword(Str s)
  {
    return keywords.get(s, false)
  }

  static const Str:Bool keywords
  static
  {
    list :=
   // TODO: this should really map to a bitmask per language
    [ "abstract",   "as",          "assert",      "boolean",
      "break",      "case",        "catch",       "class",
      "const",      "continue",    "default",     "delegate",
      "do",         "else",        "enum",        "event",
      "explicit",   "extends",     "extern",      "false",
      "final",      "finally",     "fixed",       "for",
      "foreach",    "goto",        "if",          "implements",
      "implicit",   "import",      "instanceof",  "interface",
      "internal",   "is",          "lock",        "mixin",
      "native",     "new",         "null",        "object",
      "operator",   "override",    "package",     "private",
      "protected",  "public",      "readonly",    "return",
      "sealed",     "sizeof",      "static",      "strictfp",
      "struct",     "super",       "switch",      "synchronized",
      "this",       "throw",       "throws",      "transient",
      "true",       "try",         "typeof",      "unchecked",
      "unsafe",     "using",       "virtual",     "void",
      "volatile",   "while"]

     map := Str:Bool[:]
     list.each |Str s| { map[s] = true }
     keywords = map.toImmutable
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  File srcFile
}
