//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Apr 06  Brian Frank  Creation
//
package fan.sys;

import java.io.*;
import java.security.*;

/**
 * Buf
 */
public class Buf
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public static Buf make() { return new Buf().init(DefaultCapacity); }
  public static Buf make(int capacity) { return new Buf().init(capacity); }
  public static Buf make(Int capacity) { return new Buf().init((int)capacity.val); }
  public static void make$(Buf self) { self.init(DefaultCapacity ); }
  public static void make$(Buf self, Int capacity) { self.init((int)capacity.val); }

  public static Buf make(byte[] bytes)
  {
    Buf buf  = new Buf();
    buf.buf  = bytes;
    buf.pos  = 0;
    buf.size = bytes.length;
    return buf;
  }

  public static Buf make(byte[] bytes, int size)
  {
    Buf buf  = new Buf();
    buf.buf  = bytes;
    buf.pos  = 0;
    buf.size = size;
    return buf;
  }

  public Buf init(int capacity)
  {
    this.buf  = new byte[capacity];
    this.pos  = 0;
    this.size = 0;
    return this;
  }

  static final int DefaultCapacity = 1024;

//////////////////////////////////////////////////////////////////////////
// Obj
//////////////////////////////////////////////////////////////////////////

  public Bool equals(Obj that)
  {
    if (that instanceof Buf)
    {
      Buf x = (Buf)that;
      if (size != x.size) return Bool.False;
      for (int i=0; i<size; ++i)
        if (buf[i] != x.buf[i]) return Bool.False;
      return Bool.True;
    }
    return Bool.False;
  }

  public Int hash()
  {
    int hash = 30311;
    for (int i=0; i<size; ++i)
      hash ^= buf[i] << i%32;
    return Int.make(hash);
  }

  public Str toStr()
  {
    StringBuilder s = new StringBuilder();
    s.append("0x[");
    for (int i=0; i<size; ++i)
    {
      int b = buf[i] & 0xFF;
      if (b < 0x10) s.append('0');
      s.append(Integer.toHexString(b));
      if (i > 100) { s.append("..."); break; }
      if ((i+1) % 4 == 0 && i+1<size) s.append(' ');
    }
    s.append(']');
    return Str.make(s.toString());
  }

  public Type type() { return Sys.BufType; }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public Bool empty() { return size == 0 ? Bool.True : Bool.False; }

  public final Int capacity()
  {
    return Int.pos(buf.length);
  }

  public final void capacity(Int c)
  {
    int newCapacity = (int)c.val;
    if (newCapacity < size) throw ArgErr.make("capacity < size").val;
    byte[] temp = new byte[newCapacity];
    System.arraycopy(buf, 0, temp, 0, newCapacity);
    buf = temp;
  }

  public Int size()
  {
    return Int.make(size);
  }

  public void size(Int s)
  {
    int newSize = (int)s.val;
    if (newSize > buf.length)
    {
      byte[] temp = new byte[newSize];
      System.arraycopy(buf, 0, temp, 0, buf.length);
      buf  = temp;
    }
    size = newSize;
  }

  public Int pos() { return Int.make(pos); }

  public Int remaining() { return Int.make(size-pos); }

  public Bool more() { return size-pos > 0 ? Bool.True : Bool.False; }

  public Buf seek(Int pos)
  {
    int p = (int)pos.val;
    if (p < 0) p = size + p;
    if (p < 0 || p > size) throw IndexErr.make(pos).val;
    this.pos = p;
    return this;
  }

  public Buf flip()
  {
    this.size = pos;
    this.pos  = 0;
    return this;
  }

  public Int get(Int pos)
  {
    int p = (int)pos.val;
    if (p < 0) p = size + p;
    if (p < 0 || p >= size) throw IndexErr.make(pos).val;
    return Int.make(buf[p] & 0xFF);
  }

  public Buf slice(Range range)
  {
    int s = range.start(size);
    int e = range.end(size);
    int n = e - s + 1;
    if (n < 0) throw IndexErr.make(range).val;

    byte[] slice = new byte[n];
    System.arraycopy(buf, s, slice, 0, n);

    Buf result = new Buf();
    result.buf = slice;
    result.size = n;
    result.charset(charset());
    return result;
  }

//////////////////////////////////////////////////////////////////////////
// Modification
//////////////////////////////////////////////////////////////////////////

  public Buf set(Int pos, Int b)
  {
    int p = (int)pos.val;
    if (p < 0) p = size + p;
    if (p < 0 || p >= size) throw IndexErr.make(pos).val;
    buf[p] = (byte)b.val;
    return this;
  }

  public Buf trim()
  {
    if (size == buf.length) return this;
    byte[] temp = new byte[size];
    System.arraycopy(buf, 0, temp, 0, size);
    this.buf = temp;
    return this;
  }

  public Buf clear()
  {
    this.size = 0;
    this.pos  = 0;
    return this;
  }

  public Charset charset()
  {
    return out.charset();
  }

  public void charset(Charset charset)
  {
    out.charset(charset);
    in.charset(charset);
  }

//////////////////////////////////////////////////////////////////////////
// OutStream
//////////////////////////////////////////////////////////////////////////

  public OutStream out() { return out; }

  public Buf write(Int b) { out.write(b); return this; }

  public Buf writeBuf(Buf other) { out.writeBuf(other); return this; }
  public Buf writeBuf(Buf other, Int n) { out.writeBuf(other, n); return this; }

  public Buf writeI2(Int x) { out.writeI2(x); return this; }

  public Buf writeI4(Int x) { out.writeI4(x); return this; }

  public Buf writeI8(Int x) { out.writeI8(x); return this; }

  public Buf writeF4(Float x) { out.writeF4(x); return this; }

  public Buf writeF8(Float x) { out.writeF8(x); return this; }

  public Buf writeBool(Bool x) { out.writeBool(x); return this; }

  public Buf writeUtf(Str x) { out.writeUtf(x); return this; }

  public Buf writeChar(Int c) { out.writeChar(c); return this; }

  public Buf writeChars(Str s) { out.writeChars(s); return this; }
  public Buf writeChars(Str s, Int off) { out.writeChars(s, off); return this; }
  public Buf writeChars(Str s, Int off, Int len) { out.writeChars(s, off, len); return this; }

  public Buf print(Obj obj) { out.print(obj); return this; }

  public Buf printLine() { out.printLine(); return this; }
  public Buf printLine(Obj obj) { out.printLine(obj); return this; }

  public Buf writeObj(Obj obj) { out.writeObj(obj); return this; }
  public Buf writeObj(Obj obj, Map opt) { out.writeObj(obj, opt); return this; }

//////////////////////////////////////////////////////////////////////////
// InStream
//////////////////////////////////////////////////////////////////////////

  public InStream in() { return in; }

  public Int read() {  return in.read(); }

  public Int readBuf(Buf other, Int n) { return in.readBuf(other, n); }

  public Buf unread(Int n) { in.unread(n); return this; }

  public Buf readBufFully(Buf buf, Int n) { return in.readBufFully(buf, n); }

  public Buf readAllBuf() { return in.readAllBuf(); }

  public Int peek() { return in.peek(); }

  public Int readU1() { return in.readU1(); }

  public Int readS1() { return in.readS1(); }

  public Int readU2() { return in.readU2(); }

  public Int readS2() { return in.readS2(); }

  public Int readU4() { return in.readU4(); }

  public Int readS4() { return in.readS4(); }

  public Int readS8() { return in.readS8(); }

  public Float readF4() { return in.readF4(); }

  public Float readF8() { return in.readF8(); }

  public Bool readBool() { return in.readBool(); }

  public Str readUtf() { return in.readUtf(); }

  public Int readChar() { return in.readChar(); }

  public Buf unreadChar(Int c) { in.unreadChar(c); return this; }

  public Int peekChar() { return in.peekChar(); }

  public Str readLine() { return in.readLine(); }
  public Str readLine(Int max) { return in.readLine(max); }

  public Str readStrToken() { return in.readStrToken(); }
  public Str readStrToken(Int max) { return in.readStrToken(max); }
  public Str readStrToken(Int max, Func f) { return in.readStrToken(Int.Chunk, f); }

  public List readAllLines() { return in.readAllLines(); }

  public void eachLine(Func f) { in.eachLine(f); }

  public Str readAllStr() { return in.readAllStr(); }
  public Str readAllStr(Bool normalizeNewlines)  { return in.readAllStr(normalizeNewlines); }

  public Obj readObj() { return in.readObj(); }

//////////////////////////////////////////////////////////////////////////
// Hex
//////////////////////////////////////////////////////////////////////////

  public Str toHex()
  {
    byte[] buf = this.buf;
    int size = this.size;
    char[] hexChars = Buf.hexChars;
    StringBuilder s = new StringBuilder(size*2);
    for (int i=0; i<size; ++i)
    {
      int b = buf[i] & 0xFF;
      s.append(hexChars[b>>4]).append(hexChars[b&0xf]);
    }
    return Str.make(s.toString());
  }

  public static Buf fromHex(Str str)
  {
    String s = str.val;
    int slen = s.length();
    byte[] buf = new byte[slen/2];
    int[] hexInv = Buf.hexInv;
    int size = 0;

    for (int i=0; i<slen; ++i)
    {
      int c0 = s.charAt(i);
      int n0 = c0 < 128 ? hexInv[c0] : -1;
      if (n0 < 0) continue;

      int n1 = -1;
      if (++i < slen)
      {
        int c1 = s.charAt(i);
        n1 = c1 < 128 ? hexInv[c1] : -1;
      }
      if (n1 < 0) throw IOErr.make("Invalid hex str").val;

      buf[size++] = (byte)((n0 << 4) | n1);
    }

    return Buf.make(buf, size);
  }

  private static char[] hexChars = "0123456789abcdef".toCharArray();
  private static int[] hexInv    = new int[128];
  static
  {
    for (int i=0; i<hexInv.length; ++i) hexInv[i] = -1;
    for (int i=0; i<10; ++i)  hexInv['0'+i] = i;
    for (int i=10; i<16; ++i) hexInv['a'+i-10] = hexInv['A'+i-10] = i;
  }

//////////////////////////////////////////////////////////////////////////
// Base64
//////////////////////////////////////////////////////////////////////////

  public Str toBase64()
  {
    byte[] buf = this.buf;
    int size = this.size;
    StringBuilder s = new StringBuilder();
    char[] base64chars = Buf.base64chars;
    int i = 0;

    // append full 24-bit chunks
    int end = size-2;
    for (; i<size-2; i += 3)
    {
      int c = buf[i];
      int n = (buf[i] << 16) + ((buf[i+1]) << 8) + buf[i+2];
      s.append(base64chars[(n >>> 18) & 0x3f]);
      s.append(base64chars[(n >>> 12) & 0x3f]);
      s.append(base64chars[(n >>> 6) & 0x3f]);
      s.append(base64chars[n & 0x3f]);
    }

    // pad and encode remaining bits
    int rem = size - i;
    if (rem > 0)
    {
      int n = ((buf[i] & 0xff) << 10) | (rem == 2 ? ((buf[size-1] & 0xff) << 2) : 0);
      s.append(base64chars[(n >>> 12) & 0x3f]);
      s.append(base64chars[(n >>> 6) & 0x3f]);
      s.append(rem == 2 ? base64chars[n & 0x3f] : '=');
      s.append('=');
    }

    return Str.make(s.toString());
  }

  public static Buf fromBase64(Str str)
  {
    String s = str.val;
    int slen = s.length();
    int si = 0;
    int max = slen * 6 / 8;
    byte[] buf = new byte[max];
    int size = 0;

    while (si < slen)
    {
      int n = 0;
      int v = 0;
      for (int j=0; j<4 && si<slen;)
      {
        int ch = s.charAt(si++);
        int c = ch < 128 ? base64inv[ch] : -1;
        if (c >= 0)
        {
          n |= c << (18 - j++ * 6);
          if (ch != '=') v++;
        }
      }

      if (v > 1) buf[size++] = (byte)(n >> 16);
      if (v > 2) buf[size++] = (byte)(n >> 8);
      if (v > 3) buf[size++] = (byte)n;
    }

    return Buf.make(buf, size);
  }

  private static char[] base64chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".toCharArray();
  private static final int[] base64inv = new int[128];
  static
  {
    for (int i=0; i<base64inv.length; ++i)   base64inv[i] = -1;
    for (int i=0; i<base64chars.length; ++i) base64inv[base64chars[i]] = i;
    base64inv['='] = 0;
  }

//////////////////////////////////////////////////////////////////////////
// Digest
//////////////////////////////////////////////////////////////////////////

  public Buf toDigest(Str algorithm)
  {
    try
    {
      MessageDigest md = MessageDigest.getInstance(algorithm.val);
      md.update(buf, 0, size);
      return Buf.make(md.digest());
    }
    catch (NoSuchAlgorithmException e)
    {
      throw ArgErr.make("Unknown digest algorthm: " + algorithm.val).val;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  public byte[] bytes()
  {
    byte[] r = new byte[size];
    System.arraycopy(buf, 0, r, 0, size);
    return r;
  }

  public void grow(int capacity)
  {
    if (buf.length >= capacity) return;
    byte[] temp = new byte[Math.max(capacity, size*2)];
    System.arraycopy(buf, 0, temp, 0, size);
    buf = temp;
  }

//////////////////////////////////////////////////////////////////////////
// BufOutStream
//////////////////////////////////////////////////////////////////////////

  class BufOutStream extends OutStream
  {
    public final OutStream write(Int v) { return w((int)v.val); }
    public final OutStream w(int v)
    {
      if (pos+1 >= buf.length) grow(pos+1);
      buf[pos++] = (byte)v;
      if (pos > size) size = pos;
      return this;
    }

    public OutStream writeBuf(Buf other, Int n)
    {
      int len = (int)n.val;
      if (other.pos + len > other.size) throw IOErr.make("Not enough bytes to write").val;
      grow(pos+len);
      System.arraycopy(other.buf, other.pos, buf, pos, len);
      pos += len;
      if (pos > size) size = pos;
      other.pos += len;
      return this;
    }
  }

//////////////////////////////////////////////////////////////////////////
// BufInStream
//////////////////////////////////////////////////////////////////////////

  class BufInStream extends InStream
  {
    public Int read() { int n = r(); return n < 0 ? null : Int.pos(n); }
    public int r()
    {
      if (pos >= size) return -1;
      return buf[pos++] & 0xFF;
    }

    public Int readBuf(Buf other, Int n)
    {
      if (pos >= size) return null;
      int len = Math.min(size-pos, (int)n.val);
      other.grow(other.pos+len);
      System.arraycopy(buf, pos, other.buf, other.pos, len);
      pos += len;
      other.pos += len;
      other.size += len;
      return Int.pos(len);
    }

    public InStream unread(Int n) { return unread((int)n.val); }
    public InStream unread(int n)
    {
      // unreading a buffer is a bit weird - the typical case
      // is that we are pushing back the byte we just read in
      // which case we can just rewind the position; however
      // if we pushing back a different byte then we need
      // to shift the entire buffer and insert the byte
      if (pos > 0 && buf[pos-1] == (byte)n)
      {
        pos--;
      }
      else
      {
        if (size+1 >= buf.length) grow(size+1);
        System.arraycopy(buf, pos, buf, pos+1, size);
        buf[pos] = (byte)n;
        size++;
      }
      return this;
    }

  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public byte[] buf;
  public int pos;
  public int size;
  public final OutStream out = new BufOutStream();
  public final InStream in = new BufInStream();

}