//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Mar 06  Brian Frank  Creation
//
package fan.sys;

/**
 * File represents a file or directory in a file system.
 */
public abstract class File
  extends Resource
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static File make(Uri uri)
  {
    return new LocalFile(uri);
  }

  public static File os(Str osPath)
  {
    return new LocalFile(new java.io.File(osPath.val));
  }

  public static File createTemp() { return createTemp(null, null, null); }
  public static File createTemp(Str prefix) { return createTemp(prefix, null, null); }
  public static File createTemp(Str prefix, Str suffix) { return createTemp(prefix, suffix, null); }
  public static File createTemp(Str prefix, Str suffix, File dir)
  {
    if (prefix == null || prefix.val.length() == 0) prefix = Str.make("fan");
    if (prefix.val.length() == 1) prefix = Str.make(prefix.val + "xx");
    if (prefix.val.length() == 2) prefix = Str.make(prefix.val + "x");

    if (suffix == null) suffix = Str.make(".tmp");

    java.io.File d = null;
    if (dir != null)
    {
      if (!(dir instanceof LocalFile)) throw IOErr.make("Dir is not on local file system: " + dir).val;
      d = ((LocalFile)dir).file;
    }

    try
    {
      return new LocalFile(java.io.File.createTempFile(prefix.val, suffix.val, d));
    }
    catch (java.io.IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  protected File(Uri uri)
  {
    super(uri);
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Bool equals(Obj obj)
  {
    if (obj instanceof File)
    {
      return uri.equals(((File)obj).uri);
    }
    return Bool.False;
  }

  public final int hashCode() { return uri.hashCode(); }

  public final Int hash() { return uri.hash(); }

  public Type type() { return Sys.FileType; }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public abstract Bool exists();

  public abstract Int size();

  public abstract DateTime modified();
  public abstract void modified(DateTime time);

  public abstract Str osPath();

  public abstract File parent();

  public abstract List list();

  public abstract File normalize();

  public File plus(String uri) { return plus(Uri.fromStr(uri)); }
  public abstract File plus(Uri uri);

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  public abstract Resource create();

  public abstract void delete();

  public abstract File deleteOnExit();

//////////////////////////////////////////////////////////////////////////
// IO
//////////////////////////////////////////////////////////////////////////

  public final InStream in() { return in(Int.Chunk); }
  public abstract InStream in(Int bufSize);

  public final OutStream out() { return out(Bool.False, Int.Chunk); }
  public final OutStream out(Bool append) { return out(append, Int.Chunk); }
  public abstract OutStream out(Bool append, Int bufSize);

  public final Buf readAllBuf()
  {
    return in(Int.Chunk).readAllBuf();
  }

  public final List readAllLines()
  {
    return in(Int.Chunk).readAllLines();
  }

  public final void eachLine(Func f)
  {
    in(Int.Chunk).eachLine(f);
  }

  public final Str readAllStr() { return readAllStr(Bool.True); }
  public final Str readAllStr(Bool normalizeNewlines)
  {
    return in(Int.Chunk).readAllStr(normalizeNewlines);
  }

  public final Map readProps()
  {
    return in(Int.Chunk).readProps();
  }

  public final void writeProps(Map props)
  {
    out(Bool.False, Int.Chunk).writeProps(props, Bool.True);
  }

  public final Obj readObj() { return readObj(null); }
  public final Obj readObj(Map options)
  {
    InStream in = in();
    try
    {
      return in.readObj(options);
    }
    finally
    {
      in.close();
    }
  }

  public final void writeObj(Obj obj) { writeObj(obj, null); }
  public final void writeObj(Obj obj, Map options)
  {
    OutStream out = out();
    try
    {
      out.writeObj(obj, options);
    }
    finally
    {
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public static final Str sep = Str.make(java.io.File.separator);

}