//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Mar 06  Brian Frank  Creation
//
package fan.sys;

/**
 * LocalFile represents a file or directory in the local file system.
 */
public class LocalFile
  extends File
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static Uri toUri(java.io.File file)
  {
    return toUri(file, file.isDirectory());
  }

  public static Uri toUri(java.io.File file, boolean isDir)
  {
    String path = file.getPath();
    path = path.replace('\\', '/');
    if (isDir)
    {
      // add trailing slash if not present
      if (path.length() == 0 || path.charAt(path.length()-1) != '/')
        path = path + "/";
    }

    // deal with Windoze drive name
    if (path.length() > 2 && path.charAt(1) == ':' && path.charAt(0) != '/')
      path = "/" + path;

    return Uri.fromStr(path);
  }

  public LocalFile(Uri uri)
  {
    this(uri, new java.io.File(uri.pathStr().val));
  }

  public LocalFile(java.io.File file)
  {
    this(toUri(file), file);
  }

  public LocalFile(java.io.File file, boolean isDir)
  {
    this(toUri(file, isDir), file);
  }

  public LocalFile(Uri uri, java.io.File file)
  {
    super(uri);
    this.file = file;
    if (file.exists())
    {
      if (file.isDirectory())
      {
        if (!uri.isDir().val)
          throw IOErr.make("Must use trailing slash for dir: " + uri).val;
      }
      else
      {
        if (uri.isDir().val)
          throw IOErr.make("Cannot use trailing slash for file: " + uri).val;
      }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Obj
//////////////////////////////////////////////////////////////////////////

  public Type type() { return Sys.LocalFileType; }

//////////////////////////////////////////////////////////////////////////
// File
//////////////////////////////////////////////////////////////////////////

  public Bool exists()
  {
    return file.exists() ? Bool.True : Bool.False;
  }

  public Int size()
  {
    if (file.isDirectory()) return null;
    return Int.pos(file.length());
  }

  public DateTime modified()
  {
    return DateTime.java(file.lastModified());
  }

  public void modified(DateTime time)
  {
    file.setLastModified(time.java());
  }

  public Str osPath()
  {
    return Str.make(file.getPath());
  }

  public File parent()
  {
    Uri parent = uri.parent();
    if (parent == null) return null;
    return new LocalFile(parent);
  }

  public List list()
  {
    java.io.File[] list = file.listFiles();
    int len = list == null ? 0 : list.length;
    List acc = new List(Sys.FileType, len);
    for (int i=0; i<len; ++i)
    {
      java.io.File f = list[i];
      String name = f.getName();
      if (f.isDirectory()) name += "/";
      acc.add(new LocalFile(uri.plus(Uri.fromStr(name)), f));
    }
    return acc;
  }

  public File normalize()
  {
    try
    {
      return new LocalFile(file.getCanonicalFile());
    }
    catch (java.io.IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  public File plus(Uri uri)
  {
    return make(this.uri.plus(uri));
  }

//////////////////////////////////////////////////////////////////////////
// File Management
//////////////////////////////////////////////////////////////////////////

  public Resource create()
  {
    if (isDir().val)
      createDir();
    else
      createFile();
    return this;
  }

  private void createFile()
  {
    if (file.exists())
    {
      if (file.isDirectory())
        throw IOErr.make("Already exists as dir: " + file).val;
    }
    try
    {
      java.io.FileOutputStream out = new java.io.FileOutputStream(file);
      out.close();
    }
    catch (java.io.IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  private void createDir()
  {
    if (file.exists())
    {
      if (!file.isDirectory())
        throw IOErr.make("Already exists as file: " + file).val;
    }
    else
    {
      if (!file.mkdirs())
        throw IOErr.make("Cannot createDir: " + file).val;
    }
  }

  public void delete()
  {
    if (!exists().val) return;

    if (file.isDirectory())
    {
      List kids = list();
      for (int i=0; i<kids.sz(); ++i)
        ((File)kids.get(i)).delete();
    }

    if (!file.delete())
      throw IOErr.make("Cannot delete: " + file).val;
  }

  public File deleteOnExit()
  {
    file.deleteOnExit();
    return this;
  }

//////////////////////////////////////////////////////////////////////////
// IO
//////////////////////////////////////////////////////////////////////////

  public InStream in(Int bufSize)
  {
    try
    {
      return SysInStream.make(new java.io.FileInputStream(file), bufSize);
    }
    catch (java.io.IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  public OutStream out(Bool append, Int bufSize)
  {
    try
    {
      return SysOutStream.make(new java.io.FileOutputStream(file, append.val), bufSize);
    }
    catch (java.io.IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final java.io.File file;

}