//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Nov 07  Brian Frank  Creation
//
package fan.sys;

import java.util.*;

/**
 * MemResource is the type for in-memory resources via MemResourcePeer.
 */
public class MemResource
  extends Resource
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static MemResource make(Uri uri, Obj obj)
  {
    MemResource self = new MemResource();
    make$(self, uri, obj);
    return self;
  }

  public static void make$(MemResource self, Uri uri, Obj obj)
  {
    Resource.make$(self, uri);
    self.obj      = obj;
    self.needLoad = false;
  }

  public MemResource() {}

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public Type type()
  {
    return Sys.MemResourceType;
  }

  public Bool exists()
  {
    return Bool.make(master != null);
  }

//////////////////////////////////////////////////////////////////////////
// Obj
//////////////////////////////////////////////////////////////////////////

  public Obj obj()
  {
    checkWorking();
    if (needLoad)
    {
      // lazily make thread-safe copy of master obj
      synchronized (master) { obj = Resource.safe(master.obj); }
      needLoad = false;
    }
    return obj;
  }

  public void obj(Obj obj)
  {
    checkWorking();
    this.obj = obj;
    this.needLoad = false;
  }

//////////////////////////////////////////////////////////////////////////
// Children
//////////////////////////////////////////////////////////////////////////

  public List list()
  {
    checkWorking();
    if (master == null) return emptyList;

    synchronized (master)
    {
      HashMap children = master.children;
      if (children == null || children.size() == 0)
        return emptyList;

      Resource[] acc = new Resource[children.size()];
      Iterator it = children.values().iterator();
      int n = 0;
      while (it.hasNext())
        acc[n++] = asMem((Resource)it.next()).toWorking();
      return new List(Sys.ResourceType, acc);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Lifecycle
//////////////////////////////////////////////////////////////////////////

  public void refresh()
  {
    checkMountedWorking();
    synchronized (master)
    {
      this.obj = safe(master.obj);
      this.version = master.version;
    }
  }

  public Resource create()
  {
    // get my parent's uri
    Uri myUri = uri();
    Uri parentUri = myUri.parent();
    if (parentUri == null)
      throw Err.make("Invalid Uri to mount `" + myUri + "`").val;

    // resolve my parent
    Resource parent = parentUri.resolve(null, Bool.False);
    if (parent == null)
      throw Err.make("Parent `" + parentUri + "` does not exist").val;

    // check parent is instance of MemResource
    if (!(parent instanceof MemResource))
      throw Err.make("Parent `" + parentUri + "` not MemResource: " + parent.type()).val;

    asMem(parent).createChild(this);
    return this;
  }

  private void createChild(MemResource wc)
  {
    // legend:
    //   wc:  working child
    //   mc:  master child

    checkMountedWorking();

    if (wc.exists().val)
      throw Err.make("Already mounted `" + wc.uri() + "`").val;

    Str name = wc.uri().name();
    if (name == null || name.val.length() == 0)
      throw Err.make("Child must define a name: " + wc.type()).val;


    MemResource mc = wc.newResource(wc.uri());
    mc.isMaster = true;
    mc.obj  = Resource.safe(wc.obj);

    synchronized (master)
    {
      if (master.children == null)
        master.children = new HashMap(16);

      if (master.children.get(name) != null)
        throw Err.make("Duplicate name: " + name).val;

      master.children.put(name, mc);

      wc.master = mc;
    }
  }

  public void save()
  {
    checkMountedWorking();

    Obj safeObj = Resource.safe(obj);
    synchronized (master)
    {
      if (version != master.version)
        throw Err.make("Resource has been updated since last read `" + uri() + "`").val;
      master.obj = safeObj;
      version = ++master.version;
    }
  }

  public void delete()
  {
    checkMountedWorking();

    Uri uri = uri();
    Str name = uri.name();
    Uri parentUri = uri.parent();
    if (name == null || parentUri == null)
      throw UnsupportedErr.make("Cannot delete `" + uri + "`").val;

    MemResource workingParent = asMem(parentUri.resolve());
    MemResource masterParent = workingParent.master;

    // remove from parent's master
    synchronized (masterParent)
    {
      masterParent.children.remove(name);
    }

    // cleanup my master (shouldn't really matter)
    master.obj = null;
    master.version = -1;

    // cleanup myself
    this.master = null;
  }

//////////////////////////////////////////////////////////////////////////
// Resolve
//////////////////////////////////////////////////////////////////////////

  public Resource resolve(Uri uri, Int startDepth)
  {
    List path = uri.path;
    int pathDepth = path.sz();
    MemResource p = this;

    for (int i=(int)startDepth.val; i<pathDepth; ++i)
    {
      synchronized (p)
      {
        if (p.children == null) return null;
        Resource r = (Resource)p.children.get(path.get(i));
        if (r == null) return null;
        p = asMem(r);
      }
    }

    return p.toWorking();
  }

//////////////////////////////////////////////////////////////////////////
// Overrides
//////////////////////////////////////////////////////////////////////////

  /**
   * Make a new instance of a Resource for this peer type.
   */
  public MemResource newResource(Uri uri)
  {
    return (MemResource)type().make(new List(Sys.ObjType, new Obj[] { uri, null }));
  }

  /**
   * Get a working instance of the peer which is safe to pass
   * to threads to work with the resource.  The working instance
   * should be insulated from further changes to this master
   * instance (until refreshed).
   */
  public Resource toWorking()
  {
    checkMaster();
    MemResource w = newResource(this.uri());
    w.master   = this;
    w.version  = this.version;
    w.needLoad = true;
    return w;
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Get the resource as a MemResource.
   */
  public static MemResource asMem(Resource r)
  {
    try
    {
      return (MemResource)r;
    }
    catch (ClassCastException e)
    {
      throw Err.make("Not a MemResource: " + r.type()).val;
    }
  }

  /**
   * Verify that this instance is an working instance.
   */
  public void checkWorking()
  {
    if (isMaster) throw UnsupportedErr.make("Not working instance").val;
  }

  /**
   * Verify that this instance is an unmounted working instance.
   */
  public void checkUnmountedWorking()
  {
    checkWorking();
    if (master != null)
      throw UnsupportedErr.make("Resource is already mounted `" + uri() + "`").val;
  }

  /**
   * Verify that this instance is an mounted working instance.
   */
  public void checkMountedWorking()
  {
    checkWorking();
    if (master == null)
      throw UnsupportedErr.make("Resource not mounted `" + uri() + "`").val;
  }

  /**
   * Verify that this instance is an mounted master instance.
   */
  public void checkMaster()
  {
    if (!isMaster)
      throw UnsupportedErr.make("Not master instance").val;
  }

  /**
   * Trap dynamic slot access.
   */
  public Obj trap(Str name, List args)
  {
    String n = name.val;
    if (n.equals("dump")) { dump(); return null; }
    return super.trap(name, args);
  }

  /**
   * Dump the state to standard out
   */
  public void dump()
  {
    System.out.println(getClass().getName());
    System.out.println("  isMaster = " + isMaster);
    System.out.println("  uri      = " + uri());
    System.out.println("  master   = " + master);
    System.out.println("  needLoad = " + needLoad);
    System.out.println("  obj      = " + obj);
    System.out.println("  version  = " + version);
    System.out.println("  children = " + children);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  boolean isMaster;
  MemResource master;
  boolean needLoad;
  Obj obj;
  int version;
  HashMap children;

}
