//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Nov 07  Brian Frank  Creation
//
package fan.sys;

import java.util.*;

/**
 * RootResource is an internal type for managing the
 * root of the resource namespace itself.
 */
public class RootResource
  extends MemResource
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static RootResource make(Uri uri)
  {
    RootResource self = new RootResource();
    MemResource.make$(self, uri, null);
    return self;
  }

  public static RootResource makeMaster()
  {
    RootResource self = new RootResource();
    MemResource.make$(self, Uri.fromStr("/"), null);
    self.isMaster = true;
    return self;
  }

  public RootResource() {}

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public Type type()
  {
    return Sys.RootResourceType;
  }

  public Bool exists()
  {
    return Bool.True;
  }

//////////////////////////////////////////////////////////////////////////
// Obj
//////////////////////////////////////////////////////////////////////////

  public Obj obj()
  {
    return null;
  }

  public void obj(Obj obj)
  {
    throw UnsupportedErr.make("sys::RootResource.obj").val;
  }

//////////////////////////////////////////////////////////////////////////
// Children
//////////////////////////////////////////////////////////////////////////

  public List list()
  {
    // MemResource super class mounted resources
    List list = super.list().rw();

    // `/sys`
    list.add(SysResource.root());

    // Sys.appDir
    java.io.File[] files = Sys.AppDir.listFiles();
    for (int i=0; i<files.length; ++i)
    {
      java.io.File f = files[i];
      String uri = "/" + f.getName();
      if (f.isDirectory()) uri += "/";
      list.add(new LocalFile(Uri.fromStr(uri), f));
    }

    return list;
  }

  public Resource resolve(Uri uri, Int startDepth)
  {
    // sanity check
    if (startDepth != Int.Zero)
      throw new IllegalStateException();

    // check if resolving myself
    if (uri.path.sz() == 0)
      return toWorking();

    // `/sys`
    if (uri.path(0).equals("sys"))
      return SysResource.root().resolve(uri, Int.One);

    // check MemResource super class for mounted resources
    Resource r = super.resolve(uri, startDepth);
    if (r != null) return r;

    // check Sys.appDir
    java.io.File f = new java.io.File(Sys.AppDir, uri.toString().substring(1));
    if (f.exists())
    {
      if (f.isDirectory() && !uri.isDir().val)
        uri = Uri.fromStr(uri.pathStr().val + "/");
      return new LocalFile(uri, f);
    }

    // no joy
    return null;
  }

//////////////////////////////////////////////////////////////////////////
// Unsupported
//////////////////////////////////////////////////////////////////////////

  public Resource create()
  {
    throw UnsupportedErr.make("sys::RootResource.create").val;
  }

  public void save()
  {
    throw UnsupportedErr.make("sys::RootResource.save").val;
  }

  public void delete()
  {
    throw UnsupportedErr.make("sys::RootResource.delete").val;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

}
