//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Nov 07  Brian Frank  Creation
//
package fan.sys;

import java.util.*;

/**
 * SysResource is used to wrap built-in immutable system resources.
 */
public class SysResource
  extends Resource
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static SysResource root()
  {
    if (root == null)
    {
      root = make("/sys/", new NavDir())
        .add(make("/sys/logs/",    new LogsDir()))
        .add(make("/sys/threads/", new ThreadsDir()));
    }
    return root;
  }

  public static SysResource make(String uri, Dir dir)
  {
    SysResource self = new SysResource();
    self.uri = Uri.fromStr(uri);
    self.dir = dir;
    return self;
  }

  public static SysResource make(String uri, Obj obj)
  {
    if (!obj.isImmutable().val)
      throw NotImmutableErr.make(obj.type().toStr()).val;

    SysResource self = new SysResource();
    self.uri = Uri.fromStr(uri);
    self.dir = new EmptyDir();
    self.obj = obj;
    return self;
  }

  private SysResource() {}

//////////////////////////////////////////////////////////////////////////
// Resource
//////////////////////////////////////////////////////////////////////////

  public Type type() { return Sys.SysResourceType; }

  public Bool exists() { return Bool.True; }

  public Obj obj() { return obj; }

  public List list() { return dir.list(); }

  public Resource resolve(Uri uri, Int startDepth)
  {
    if (uri.path.size() == startDepth) return this;
    SysResource kid = dir.resolve((Str)uri.path().get(startDepth));
    if (kid == null) return null;
    return kid.resolve(uri, startDepth.plus(Int.One));
  }

//////////////////////////////////////////////////////////////////////////
// Dir
//////////////////////////////////////////////////////////////////////////

  static abstract class Dir
  {
    abstract List list();
    abstract SysResource resolve(Str name);
  }

//////////////////////////////////////////////////////////////////////////
// EmptyDir
//////////////////////////////////////////////////////////////////////////

  static class EmptyDir extends Dir
  {
    List list() { return Resource.emptyList; }
    SysResource resolve(Str name) { return null; }
  }

//////////////////////////////////////////////////////////////////////////
// NavDir
//////////////////////////////////////////////////////////////////////////

  SysResource add(SysResource kid)
  {
    ((NavDir)dir).add(kid);
    return this;
  }

  static class NavDir extends Dir
  {
    List list()
    {
      return list.ro();
    }

    SysResource resolve(Str name)
    {
      return (SysResource)map.get(name);
    }

    void add(SysResource kid)
    {
      map.put(kid.name(), kid);
      list.add(kid);
    }

    HashMap map = new HashMap();
    List list = new List(Sys.ResourceType);
  }

//////////////////////////////////////////////////////////////////////////
// ThreadsDir
//////////////////////////////////////////////////////////////////////////

  static class ThreadsDir extends Dir
  {
    List list()
    {
      List threads = Thread.list();
      List acc = new List(Sys.ResourceType, threads.sz());
      for (int i=0; i<threads.sz(); ++i)
      {
        Thread t = (Thread)threads.get(i);
        acc.add(make("/sys/threads/" + t.name(), t));
      }
      return acc;
    }

    SysResource resolve(Str name)
    {
      Thread t = Thread.find(name, Bool.False);
      if (t == null) return null;
      return make("/sys/threads/" + t.name(), t);
    }
  }

//////////////////////////////////////////////////////////////////////////
// LogsDir
//////////////////////////////////////////////////////////////////////////

  static class LogsDir extends Dir
  {
    List list()
    {
      List logs = Log.list();
      List acc = new List(Sys.ResourceType, logs.sz());
      for (int i=0; i<logs.sz(); ++i)
      {
        Log log = (Log)logs.get(i);
        acc.add(make("/sys/logs/" + log.name(), log));
      }
      return acc;
    }

    SysResource resolve(Str name)
    {
      Log log  = Log.find(name, Bool.False);
      if (log == null) return null;
      return make("/sys/logs/" + log.name(), log);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private static SysResource root;

  Obj obj;
  Dir dir;
}
