//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Dec 05  Brian Frank  Creation
//
package fanx.tools;

import java.io.File;
import fan.sys.*;
import fan.sys.Thread;
import fanx.util.*;

/**
 * Fan runtime command line tool.
 */
public class Fan
{

//////////////////////////////////////////////////////////////////////////
// Execute
//////////////////////////////////////////////////////////////////////////

  public int execute(String target, String[] args)
    throws Exception
  {
    // args
    for (int i=0; i<args.length; ++i)
      Sys.args.add(Str.make(args[i]));

    // first try as file name
    File file = new File(target);
    if (file.exists() && !file.isDirectory())
    {
      return executeFile(file);
    }
    else
    {
      return executeType(target);
    }
  }

  public int executeFile(File file)
    throws Exception
  {
    LocalFile f = (LocalFile)(new LocalFile(file).normalize());

    // use Fan reflection to run compiler::Main.compileScript(File)
    Pod pod = null;
    try
    {
      pod = (Pod)Slot.findMethod("compiler::Main.compileScript", true).call1(f);
    }
    catch (Err.Val e)
    {
      System.out.println("ERROR: cannot compile script");
      if (!e.getClass().getName().startsWith("fan.compiler"))
        e.err().trace();
      return -1;
    }
    catch (Exception e)
    {
      System.out.println("ERROR: cannot compile script");
      if (!e.getClass().getName().startsWith("fan.compiler"))
        e.printStackTrace();
      return -1;
    }

    List types = pod.types();
    if (types.sz() == 0)
    {
      System.out.println("ERROR: fan script file defines no classes");
      return -1;
    }

    Type type = null;
    Method main = null;
    for (int i=0; i<types.sz(); ++i)
    {
      type = (Type)pod.types().get(i);
      main = type.method("main", false);
      if (main != null) break;
    }

    if (main == null)
    {
      System.out.println("ERROR: missing main method: " + ((Type)types.get(0)).name() + ".main()");
      return -1;
    }

    try
    {
      Object result = null;
      if (main.isStatic().val)
        return toResult(main.call0());
      else
        return toResult(main.call1(type.make()));
    }
    catch (Err.Val ex)
    {
      ex.err().trace();
      return -1;
    }
  }

  public int executeType(String target)
    throws Exception
  {
    if (target.indexOf("::") < 0) target += "::Main.main";
    else if (target.indexOf('.') < 0) target += ".main";

    try
    {
      Method main = Slot.findMethod(target, true);
      if (main.isStatic().val)
        return toResult(main.call0());
      else
        return toResult(main.call1(main.parent().make()));
    }
    catch (Err.Val ex)
    {
      ex.err().trace();
      return -1;
    }
  }

  static int toResult(Object obj)
  {
    if (obj instanceof Int) return (int)((Int)obj).val;
    return 0;
  }

//////////////////////////////////////////////////////////////////////////
// Version
//////////////////////////////////////////////////////////////////////////

  static void version(String progName)
  {
    println(progName);
    println("Copyright (c) 2006-2008, Brian Frank and Andy Frank");
    println("Licensed under the Academic Free License version 3.0");
    println("");
    println("Java Runtime:");
    println("  java.vm.name:    " + System.getProperty("java.vm.name"));
    println("  java.vm.vendor:  " + System.getProperty("java.vm.vendor"));
    println("  java.vm.version: " + System.getProperty("java.vm.version"));
    println("  java.home:       " + System.getProperty("java.home"));
    println("  fan.home:        " + Sys.HomeDir);
    println("  sys.version:     " + Sys.SysPod.version());
  }

  static void pods(String progName)
  {
    version(progName);

    long t1 = System.nanoTime();
    List pods = Pod.list();
    long t2 = System.nanoTime();

    println("");
    println("Fan Pods [" + (t2-t1)/1000000L + "ms]:");

    for (int i=0; i<pods.sz(); ++i)
    {
      Pod pod = (Pod)pods.get(i);
      System.out.println("  " + pod.name().justl(Int.make(14)) + "  " + pod.version());
    }
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  public int run(String[] args)
  {
    try
    {
      if (args.length == 0) { help(); return -1; }

      // process args
      for (int i=0; i<args.length; ++i)
      {
        String a = args[i].intern();
        if (a.length() == 0) continue;
        if (a == "-help" || a == "-h" || a == "-?")
        {
          help();
          return -1;
        }
        else if (a == "-version")
        {
          version("Fan Launcher");
          return -1;
        }
        else if (a == "-pods")
        {
          pods("Fan Launcher");
          return -1;
        }
        else if (a.charAt(0) == '-')
        {
          System.out.println("WARNING: Unknown option " + a);
        }
        else
        {
          String target = a;
          String[] targetArgs = new String[args.length-i-1];
          System.arraycopy(args, i+1, targetArgs, 0, targetArgs.length);
          return execute(target, targetArgs);
        }
      }

      help();
      return -1;
    }
    catch (Throwable e)
    {
      e.printStackTrace();
      return -1;
    }
  }

  void help()
  {
    println("Fan Launcher");
    println("Usage:");
    println("  fan [options] <pod>::<type>.<method> [args]*");
    println("  fan [options] <filename> [args]*");
    println("Options:");
    println("  -help, -h, -?  print usage help");
    println("  -version       print version information");
    println("  -pods          list installed pods");
  }

  public static void println(String s)
  {
    System.out.println(s);
  }

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public static void main(final String[] args)
    throws Exception
  {
    new Thread(Str.make("main"))
    {
      public Obj run()
      {
        System.exit(new Fan().run(args));
        return null;
      }
    }.start().join();
  }

}