//
// Copyright (c) 2007 Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   14 Nov 07  Brian Frank  Creation
//
package fanx.tools;

import fan.sys.*;
import fan.sys.Thread;
import fanx.util.*;

/**
 * Fan Daemon
 */
public class Fand
{

//////////////////////////////////////////////////////////////////////////
// Process Arguments
//////////////////////////////////////////////////////////////////////////

  /**
   * Process the arguments
   * NOTE: this method must be run statically *before* the
   * Sys class is loaded, so that we have a chance to init
   * the appDir
   */
  static boolean processArgs(String[] args)
  {
    if (args.length == 0) { help(); return false; }

    String appName = null;
    for (int i=0; i<args.length; ++i)
    {
      String a = args[i].intern();
      if (a.length() == 0) continue;
      if (a == "-help" || a == "-h" || a == "-?")
      {
        help();
        return false;
      }
      else if (a == "-version")
      {
        Fan.version("Fan Daemon");
        return false;
      }
      else if (a.charAt(0) == '-')
      {
        System.out.println("WARNING: Unknown option " + a);
      }
      else
      {
        if (appName == null)
          appName = a;
        else
          println("WARNING: Argument ignored '" + a + "'");
      }
    }

    if (appName == null)
    {
      help();
      return false;
    }

    // system property for Sys.appDir
    System.getProperties().put("fan.appDir", "$home/fand/" + appName + "/");
    return true;
  }

  static void help()
  {
    println("Usage:");
    println("  fand [options] <appName>");
    println("Options:");
    println("  -help, -h, -?  print usage help");
    println("  -version       print version information");
  }

  static void println(String s)
  {
    System.out.println(s);
  }

//////////////////////////////////////////////////////////////////////////
// Boot
//////////////////////////////////////////////////////////////////////////

  public int boot()
  {
    try
    {
      if (!init()) return -1;
      if (!loadServices()) return -1;
      return waitToExit();
    }
    catch (Throwable e)
    {
      e.printStackTrace();
      return -1;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Init
//////////////////////////////////////////////////////////////////////////

  /**
   * Initialize all the directories and internal state.
   */
  boolean init()
  {
    // resolve app directory
    appDir = Sys.appDir();
    if (!appDir.exists().val)
    {
      log.error("File not found `" + appDir + "`");
      return false;
    }

    log.info("Fand version " + Sys.SysPod.version());
    log.info("App dir `" + appDir + "`");
    return true;
  }

//////////////////////////////////////////////////////////////////////////
// Load Services
//////////////////////////////////////////////////////////////////////////

  /**
   * Find all the "appDir/services/*.fog" files and
   * load them as services.
   */
  boolean loadServices()
  {
    List services = new List(Sys.ObjType);

    // load all the fog files under /appDir/services
    File dir = appDir.plus(Uri.fromStr("services/"));
    List files = dir.list();
    for (int i=0; i<files.sz(); ++i)
    {
      Thread t = loadService((File)files.get(i));
      if (t != null) services.add(t);
    }

    // if no services defined, bail
    if (services.sz() == 0)
    {
      log.error("No services defined under `" + dir + "`");
      return false;
    }

    // now start them all
    for (int i=0; i<services.sz(); ++i)
    {
      Thread t = (Thread)services.get(i);
      t.start();
    }

    return true;
  }

  /**
   * Given a file attempt to deserialize it as a Thread.
   */
  Thread loadService(File f)
  {
    try
    {
      if (f.ext() == null) return null;
      if (!f.ext().val.equals("fog")) return null;

      Map options = new Map(Sys.StrType, Sys.ObjType);
      options.add(Str.make("makeArgs"), new List(Sys.ObjType, new Obj[] { f.basename() }));
      Obj obj = f.readObj(options);

      if (!(obj instanceof Thread))
      {
        log.error("Cannot load `" + f + "`; Not a sys::Thread: " + obj.type());
        return null;
      }

      return (Thread)obj;
    }
    catch (Throwable e)
    {
      log.error("Cannot load `" + f + "`", e);
      return null;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Wait To Exit
//////////////////////////////////////////////////////////////////////////

  int waitToExit()
  {
    isAlive = true;
    while (isAlive)
    {
      try
      {
        java.lang.Thread.sleep(60000);
      }
      catch (Throwable e)
      {
        log.error("waitToExit", e);
      }
    }
    return 0;
  }

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public static void main(final String[] args)
    throws Exception
  {
    if (!processArgs(args)) System.exit(-1);
    new Thread(Str.make("main"))
    {
      public Obj run()
      {
        System.exit(new Fand().boot());
        return null;
      }
    }.start().join();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Log log = Log.get("fand");
  File appDir;
  boolean isAlive;

}