//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Feb 07  Andy Frank  Creation
//

using System;
using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// Buf
  /// </summary>
  public class Buf : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public static Buf make() { return new Buf().init(DefaultCapacity); }
    public static Buf make(int capacity) { return new Buf().init(capacity); }
    public static Buf make(Int capacity) { return new Buf().init((int)capacity.val); }
    public static void make_(Buf self) { self.init(DefaultCapacity ); }
    public static void make_(Buf self, Int capacity) { self.init((int)capacity.val); }

    public static Buf make(byte[] bytes)
    {
      Buf buf  = new Buf();
      buf.m_buf  = bytes;
      buf.m_pos  = 0;
      buf.m_size = bytes.Length;
      return buf;
    }

    public static Buf make(byte[] bytes, int size)
    {
      Buf buf  = new Buf();
      buf.m_buf  = bytes;
      buf.m_pos  = 0;
      buf.m_size = size;
      return buf;
    }

    public Buf init(int capacity)
    {
      m_buf  = new byte[capacity];
      m_pos  = 0;
      m_size = 0;
      return this;
    }

    protected Buf()
    {
      m_out = new BufOutStream(this);
      m_in = new BufInStream(this);
    }

    internal const int DefaultCapacity = 1024;

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Bool equals(Obj that)
    {
      if (that is Buf)
      {
        Buf x = (Buf)that;
        if (m_size != x.m_size) return Bool.False;
        for (int i=0; i<m_size; ++i)
          if (m_buf[i] != x.m_buf[i]) return Bool.False;
        return Bool.True;
      }
      return Bool.False;
    }

    public override Int hash()
    {
      int hash = 30311;
      for (int i=0; i<m_size; ++i)
        hash ^= m_buf[i] << i%32;
      return Int.make(hash);
    }

    public override Str toStr()
    {
      StringBuilder s = new StringBuilder();
      s.Append("0x[");
      for (int i=0; i<m_size; ++i)
      {
        int b = m_buf[i] & 0xFF;
        if (b < 0x10) s.Append('0');
        s.Append(b.ToString("X").ToLower());
        if (i > 100) { s.Append("..."); break; }
        if ((i+1) % 4 == 0 && i+1<m_size) s.Append(' ');
      }
      s.Append(']');
      return Str.make(s.ToString());
    }

    public override Type type() { return Sys.BufType; }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    public Bool empty() { return m_size == 0 ? Bool.True : Bool.False; }

    public Int capacity()
    {
      return Int.pos(m_buf.Length);
    }

    public void capacity(Int c)
    {
      int newCapacity = (int)c.val;
      if (newCapacity < m_size) throw ArgErr.make("capacity < size").val;
      byte[] temp = new byte[newCapacity];
      System.Array.Copy(m_buf, 0, temp, 0, newCapacity);
      m_buf = temp;
    }

    public Int size()
    {
      return Int.make(m_size);
    }

    public void size(Int s)
    {
      int newSize = (int)s.val;
      if (newSize > m_buf.Length)
      {
        byte[] temp = new byte[newSize];
        System.Array.Copy(m_buf, 0, temp, 0, m_buf.Length);
        m_buf = temp;
      }
      m_size = newSize;
    }

    public Int pos() { return Int.make(m_pos); }

    public Int remaining() { return Int.make(m_size-m_pos); }

    public Bool more() { return m_size-m_pos > 0 ? Bool.True : Bool.False; }

    public Buf seek(Int pos)
    {
      int p = (int)pos.val;
      if (p < 0) p = m_size + p;
      if (p < 0 || p > m_size) throw IndexErr.make(pos).val;
      m_pos = p;
      return this;
    }

    public Buf flip()
    {
      m_size = m_pos;
      m_pos  = 0;
      return this;
    }

    public Int get(Int pos)
    {
      int p = (int)pos.val;
      if (p < 0) p = m_size + p;
      if (p < 0 || p >= m_size) throw IndexErr.make(pos).val;
      return Int.make(m_buf[p] & 0xFF);
    }

    public Buf slice(Range range)
    {
      int s = range.start(m_size);
      int e = range.end(m_size);
      int n = e - s + 1;
      if (n < 0) throw IndexErr.make(range).val;

      byte[] slice = new byte[n];
      System.Array.Copy(m_buf, s, slice, 0, n);

      Buf result = new Buf();
      result.m_buf = slice;
      result.m_size = n;
      result.charset(charset());
      return result;
    }

  //////////////////////////////////////////////////////////////////////////
  // Modification
  //////////////////////////////////////////////////////////////////////////

    public Buf set(Int pos, Int b)
    {
      int p = (int)pos.val;
      if (p < 0) p = m_size + p;
      if (p < 0 || p >= m_size) throw IndexErr.make(pos).val;
      m_buf[p] = (byte)b.val;
      return this;
    }

    public Buf trim()
    {
      if (m_size == m_buf.Length) return this;
      byte[] temp = new byte[m_size];
      System.Array.Copy(m_buf, 0, temp, 0, m_size);
      m_buf = temp;
      return this;
    }

    public Buf clear()
    {
      m_size = 0;
      m_pos  = 0;
      return this;
    }

    public Charset charset()
    {
      return m_out.charset();
    }

    public void charset(Charset charset)
    {
      m_out.charset(charset);
      m_in.charset(charset);
    }

  //////////////////////////////////////////////////////////////////////////
  // OutStream
  //////////////////////////////////////////////////////////////////////////

    public OutStream @out() { return m_out; }

    public Buf write(Int b) { m_out.write(b); return this; }

    public Buf writeBuf(Buf other) { m_out.writeBuf(other); return this; }
    public Buf writeBuf(Buf other, Int n) { m_out.writeBuf(other, n); return this; }

    public Buf writeI2(Int x) { m_out.writeI2(x); return this; }

    public Buf writeI4(Int x) { m_out.writeI4(x); return this; }

    public Buf writeI8(Int x) { m_out.writeI8(x); return this; }

    public Buf writeF4(Float x) { m_out.writeF4(x); return this; }

    public Buf writeF8(Float x) { m_out.writeF8(x); return this; }

    public Buf writeBool(Bool x) { m_out.writeBool(x); return this; }

    public Buf writeUtf(Str x) { m_out.writeUtf(x); return this; }

    public Buf writeChar(Int c) { m_out.writeChar(c); return this; }

    public Buf writeChars(Str s) { m_out.writeChars(s); return this; }
    public Buf writeChars(Str s, Int off) { m_out.writeChars(s, off); return this; }
    public Buf writeChars(Str s, Int off, Int len) { m_out.writeChars(s, off, len); return this; }

    public Buf print(Obj obj) { m_out.print(obj); return this; }

    public Buf printLine() { m_out.printLine(); return this; }
    public Buf printLine(Obj obj) { m_out.printLine(obj); return this; }

    public Buf writeObj(Obj obj) { m_out.writeObj(obj); return this; }
    public Buf writeObj(Obj obj, Map opt) { m_out.writeObj(obj, opt); return this; }

  //////////////////////////////////////////////////////////////////////////
  // InStream
  //////////////////////////////////////////////////////////////////////////

    public InStream @in() { return m_in; }

    public Int read() {  return m_in.read(); }

    public Int readBuf(Buf other, Int n) { return m_in.readBuf(other, n); }

    public Buf unread(Int n) { m_in.unread(n); return this; }

    public Buf readBufFully(Buf buf, Int n) { return m_in.readBufFully(buf, n); }

    public Buf readAllBuf() { return m_in.readAllBuf(); }

    public Int peek() { return m_in.peek(); }

    public Int readU1() { return m_in.readU1(); }

    public Int readS1() { return m_in.readS1(); }

    public Int readU2() { return m_in.readU2(); }

    public Int readS2() { return m_in.readS2(); }

    public Int readU4() { return m_in.readU4(); }

    public Int readS4() { return m_in.readS4(); }

    public Int readS8() { return m_in.readS8(); }

    public Float readF4() { return m_in.readF4(); }

    public Float readF8() { return m_in.readF8(); }

    public Bool readBool() { return m_in.readBool(); }

    public Str readUtf() { return m_in.readUtf(); }

    public Int readChar() { return m_in.readChar(); }

    public Buf unreadChar(Int c) { m_in.unreadChar(c); return this; }

    public Int peekChar() { return m_in.peekChar(); }

    public Str readLine() { return m_in.readLine(); }
    public Str readLine(Int max) { return m_in.readLine(max); }

    public Str readStrToken() { return m_in.readStrToken(); }
    public Str readStrToken(Int max) { return m_in.readStrToken(max); }
    public Str readStrToken(Int max, Func f) { return m_in.readStrToken(Int.Chunk, f); }

    public List readAllLines() { return m_in.readAllLines(); }

    public void eachLine(Func f) { m_in.eachLine(f); }

    public Str readAllStr() { return m_in.readAllStr(); }
    public Str readAllStr(Bool normalizeNewlines)  { return m_in.readAllStr(normalizeNewlines); }

    public Obj readObj() { return m_in.readObj(); }

  //////////////////////////////////////////////////////////////////////////
  // Hex
  //////////////////////////////////////////////////////////////////////////

    public Str toHex()
    {
      byte[] buf = m_buf;
      int size = m_size;
      char[] hexChars = Buf.hexChars;
      StringBuilder s = new StringBuilder(size*2);
      for (int i=0; i<size; ++i)
      {
        int b = buf[i] & 0xFF;
        s.Append(hexChars[b>>4]).Append(hexChars[b&0xf]);
      }
      return Str.make(s.ToString());
    }

    public static Buf fromHex(Str str)
    {
      string s = str.val;
      int slen = s.Length;
      byte[] buf = new byte[slen/2];
      int[] hexInv = Buf.hexInv;
      int size = 0;

      for (int i=0; i<slen; ++i)
      {
        int c0 = s[i];
        int n0 = c0 < 128 ? hexInv[c0] : -1;
        if (n0 < 0) continue;

        int n1 = -1;
        if (++i < slen)
        {
          int c1 = s[i];
          n1 = c1 < 128 ? hexInv[c1] : -1;
        }
        if (n1 < 0) throw IOErr.make("Invalid hex str").val;

        buf[size++] = (byte)((n0 << 4) | n1);
      }

      return Buf.make(buf, size);
    }

  //////////////////////////////////////////////////////////////////////////
  // Base64
  //////////////////////////////////////////////////////////////////////////

    public Str toBase64()
    {
      byte[] buf = m_buf;
      int size = m_size;
      StringBuilder s = new StringBuilder();
      char[] base64chars = Buf.base64chars;
      int i = 0;

      // Append full 24-bit chunks
      int end = size-2;
      for (; i<size-2; i += 3)
      {
        int c = buf[i];
        int n = (buf[i] << 16) + ((buf[i+1]) << 8) + buf[i+2];
        s.Append(base64chars[(n >> 18) & 0x3f]);
        s.Append(base64chars[(n >> 12) & 0x3f]);
        s.Append(base64chars[(n >> 6) & 0x3f]);
        s.Append(base64chars[n & 0x3f]);
      }

      // pad and encode remaining bits
      int rem = size - i;
      if (rem > 0)
      {
        int n = ((buf[i] & 0xff) << 10) | (rem == 2 ? ((buf[size-1] & 0xff) << 2) : 0);
        s.Append(base64chars[(n >> 12) & 0x3f]);
        s.Append(base64chars[(n >> 6) & 0x3f]);
        s.Append(rem == 2 ? base64chars[n & 0x3f] : '=');
        s.Append('=');
      }

      return Str.make(s.ToString());
    }

    public static Buf fromBase64(Str str)
    {
      string s = str.val;
      int slen = s.Length;
      int si = 0;
      int max = slen * 6 / 8;
      byte[] buf = new byte[max];
      int size = 0;

      while (si < slen)
      {
        int n = 0;
        int v = 0;
        for (int j=0; j<4 && si<slen;)
        {
          int ch = s[si++];
          int c = ch < 128 ? base64inv[ch] : -1;
          if (c >= 0)
          {
            n |= c << (18 - j++ * 6);
            if (ch != '=') v++;
          }
        }

        if (v > 1) buf[size++] = (byte)(n >> 16);
        if (v > 2) buf[size++] = (byte)(n >> 8);
        if (v > 3) buf[size++] = (byte)n;
      }

      return Buf.make(buf, size);
    }

  //////////////////////////////////////////////////////////////////////////
  // Digest
  //////////////////////////////////////////////////////////////////////////

    public Buf toDigest(Str algorithm)
    {
      string alg = algorithm.val;
      if (alg == "SHA-1") alg = "SHA1";  // to make .NET happy
      HashAlgorithm ha = HashAlgorithm.Create(alg);
      if (ha == null)
        throw ArgErr.make("Unknown digest algorthm: " + algorithm.val).val;
      return Buf.make(ha.ComputeHash(m_buf, 0, m_size));
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    public byte[] bytes()
    {
      byte[] r = new byte[m_size];
      System.Array.Copy(m_buf, 0, r, 0, m_size);
      return r;
    }

    public void grow(int capacity)
    {
      if (m_buf.Length >= capacity) return;
      byte[] temp = new byte[Math.Max(capacity, m_size*2)];
      System.Array.Copy(m_buf, 0, temp, 0, m_size);
      m_buf = temp;
    }

  //////////////////////////////////////////////////////////////////////////
  // BufOutStream
  //////////////////////////////////////////////////////////////////////////

    class BufOutStream : OutStream
    {
      Buf p; // parent
      public BufOutStream(Buf p) { this.p = p; }
      public override OutStream write(Int v) { return w((int)v.val); }
      public override OutStream w(int v)
      {
        if (p.m_pos+1 >= p.m_buf.Length) p.grow(p.m_pos+1);
        p.m_buf[p.m_pos++] = (byte)v;
        if (p.m_pos > p.m_size) p.m_size = p.m_pos;
        return this;
      }

      public override OutStream writeBuf(Buf other, Int n)
      {
        int len = (int)n.val;
        if (other.m_pos + len > other.m_size) throw IOErr.make("Not enough bytes to write").val;
        p.grow(p.m_pos+len);
        System.Array.Copy(other.m_buf, other.m_pos, p.m_buf, p.m_pos, len);
        p.m_pos += len;
        if (p.m_pos > p.m_size) p.m_size = p.m_pos;
        other.m_pos += len;
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // BufInStream
  //////////////////////////////////////////////////////////////////////////

    class BufInStream : InStream
    {
      Buf p; // parent
      public BufInStream(Buf p) { this.p = p; }
      public override Int read() { int n = r(); return n < 0 ? null : Int.pos(n); }
      public override int r()
      {
        if (p.m_pos >= p.m_size) return -1;
        return p.m_buf[p.m_pos++] & 0xFF;
      }

      public override Int readBuf(Buf other, Int n)
      {
        if (p.m_pos >= p.m_size) return null;
        int len = Math.Min(p.m_size-p.m_pos, (int)n.val);
        other.grow(other.m_pos+len);
        System.Array.Copy(p.m_buf, p.m_pos, other.m_buf, other.m_pos, len);
        p.m_pos += len;
        other.m_pos += len;
        other.m_size += len;
        return Int.pos(len);
      }

      public override InStream unread(Int n) { return unread((int)n.val); }
      public override InStream unread(int n)
      {
        // unreading a buffer is a bit weird - the typical case
        // is that we are pushing back the byte we just read m_in
        // which case we can just rewind the position; however
        // if we pushing back a different byte then we need
        // to shift the entire buffer and insert the byte
        if (p.m_pos > 0 && p.m_buf[p.m_pos-1] == (byte)n)
        {
          p.m_pos--;
        }
        else
        {
          if (p.m_size+1 >= p.m_buf.Length) p.grow(p.m_size+1);
          System.Array.Copy(p.m_buf, p.m_pos, p.m_buf, p.m_pos+1, p.m_size);
          p.m_buf[p.m_pos] = (byte)n;
          p.m_size++;
        }
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Static
  //////////////////////////////////////////////////////////////////////////

    private static readonly char[] hexChars = "0123456789abcdef".ToCharArray();
    private static readonly int[] hexInv    = new int[128];

    private static readonly char[] base64chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".ToCharArray();
    private static readonly int[] base64inv = new int[128];

    static Buf()
    {
      // base64
      for (int i=0; i<hexInv.Length; ++i) hexInv[i] = -1;
      for (int i=0; i<10; ++i)  hexInv['0'+i] = i;
      for (int i=10; i<16; ++i) hexInv['a'+i-10] = hexInv['A'+i-10] = i;

      // hex
      for (int i=0; i<base64inv.Length; ++i)   base64inv[i] = -1;
      for (int i=0; i<base64chars.Length; ++i) base64inv[base64chars[i]] = i;
      base64inv['='] = 0;
    }


  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public byte[] m_buf;
    public int m_pos;
    public int m_size;
    public readonly OutStream m_out;
    public readonly InStream m_in;

  }
}