//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Jan 07  Andy Frank  Creation
//

namespace Fan.Sys
{
  /// <summary>
  /// File represents a file or directory in a file system.
  /// <summary>
  public abstract class File : Resource
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static File make(Uri uri)
    {
      return new LocalFile(uri);
    }

    public static File os(Str osPath)
    {
      System.IO.FileSystemInfo info = (System.IO.Directory.Exists(osPath.val))
        ? new System.IO.DirectoryInfo(osPath.val) as System.IO.FileSystemInfo
        : new System.IO.FileInfo(osPath.val) as System.IO.FileSystemInfo;
      return new LocalFile(osPath.val, info);
    }

    public static File createTemp() { return createTemp(null, null, null); }
    public static File createTemp(Str prefix) { return createTemp(prefix, null, null); }
    public static File createTemp(Str prefix, Str suffix) { return createTemp(prefix, suffix, null); }
    public static File createTemp(Str prefix, Str suffix, File dir)
    {
      if (prefix == null || prefix.val.Length == 0) prefix = Str.make("fan");
      if (suffix == null) suffix = Str.make(".tmp");

      string parent = null;
      if (dir == null)
      {
        parent = System.IO.Path.GetTempPath();
      }
      else
      {
        if (!(dir is LocalFile)) throw IOErr.make("Dir is not on local file system: " + dir).val;
        parent = ((LocalFile)dir).m_file.FullName;
      }

      try
      {
        string name = parent + '\\' + prefix.val + suffix.val;
        int count = 1;
        while (System.IO.File.Exists(name))
          name = parent + '\\' + prefix.val + (count++) + suffix.val;
        LocalFile temp = new LocalFile(new System.IO.FileInfo(name));
        temp.create();
        return temp;
      }
      catch (System.IO.IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    protected File(Uri uri)
      : base(uri)
    {
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Bool equals(Obj obj)
    {
      if (obj is File)
      {
        return m_uri.equals(((File)obj).m_uri);
      }
      return Bool.False;
    }

    public override int GetHashCode() { return m_uri.GetHashCode(); }

    public override Int hash() { return m_uri.hash(); }

    public override Type type() { return Sys.FileType; }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    public abstract override Bool exists();

    public abstract Int size();

    public abstract DateTime modified();
    public abstract void modified(DateTime time);

    public abstract Str osPath();

    public abstract File parent();

    public abstract override List list();

    public abstract File normalize();

    public File plus(string uri) { return plus(Uri.fromStr(uri)); }
    public abstract File plus(Uri uri);

  //////////////////////////////////////////////////////////////////////////
  // Management
  //////////////////////////////////////////////////////////////////////////

    public abstract override Resource create();

    public abstract override void delete();

    public abstract File deleteOnExit();

  //////////////////////////////////////////////////////////////////////////
  // IO
  //////////////////////////////////////////////////////////////////////////

    public InStream @in() { return @in(defaultBufSize); }
    public abstract InStream @in(Int bufSize);

    public OutStream @out() { return @out(Bool.False, defaultBufSize); }
    public OutStream @out(Bool append) { return @out(append, defaultBufSize); }
    public abstract OutStream @out(Bool append, Int bufSize);

    private static readonly Int defaultBufSize = Int.make(4096);

    public Buf readAllBuf()
    {
      return @in(defaultBufSize).readAllBuf();
    }

    public List readAllLines()
    {
      return @in(defaultBufSize).readAllLines();
    }

    public void eachLine(Func f)
    {
      @in(Int.Chunk).eachLine(f);
    }

    public Str readAllStr() { return readAllStr(Bool.True); }
    public Str readAllStr(Bool normalizeNewlines)
    {
      return @in(defaultBufSize).readAllStr(normalizeNewlines);
    }

    public Map readProps()
    {
      return @in(defaultBufSize).readProps();
    }

    public void writeProps(Map props)
    {
      @out(Bool.False, defaultBufSize).writeProps(props, Bool.True);
    }

    public Obj readObj() { return readObj(null); }
    public Obj readObj(Map options)
    {
      InStream ins = @in();
      try
      {
        return ins.readObj(options);
      }
      finally
      {
        ins.close();
      }
    }

    public void writeObj(Obj obj) { writeObj(obj, null); }
    public void writeObj(Obj obj, Map options)
    {
      OutStream outs = @out();
      try
      {
        outs.writeObj(obj, options);
      }
      finally
      {
        outs.close();
      }
    }

  }
}