//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 Nov 07  Andy Frank  Creation
//

using System.Collections;

namespace Fan.Sys
{

  /// <summary>
  /// MemResource is the type for in-memory resources via MemResourcePeer.
  /// </summary>
  public class MemResource : Resource
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static MemResource make(Uri uri, Obj obj)
    {
      MemResource self = new MemResource();
      make_(self, uri, obj);
      return self;
    }

    public static void make_(MemResource self, Uri uri, Obj obj)
    {
      Resource.make_(self, uri);
      self.m_obj      = obj;
      self.m_needLoad = false;
    }

    public MemResource() {}

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Type type()
    {
      return Sys.MemResourceType;
    }

    public override Bool exists()
    {
      return Bool.make(m_master != null);
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Obj obj()
    {
      checkWorking();
      if (m_needLoad)
      {
        // lazily make thread-safe copy of master obj
        lock (m_master) { m_obj = Resource.safe(m_master.m_obj); }
        m_needLoad = false;
      }
      return m_obj;
    }

    public override void obj(Obj obj)
    {
      checkWorking();
      this.m_obj = obj;
      this.m_needLoad = false;
    }

  //////////////////////////////////////////////////////////////////////////
  // Children
  //////////////////////////////////////////////////////////////////////////

    public override List list()
    {
      checkWorking();
      if (m_master == null) return emptyList;

      lock (m_master)
      {
        Hashtable children = m_master.m_children;
        if (children == null || children.Count == 0)
          return emptyList;

        Resource[] acc = new Resource[children.Count];
        IDictionaryEnumerator en = children.GetEnumerator();
        int n = 0;
        while (en.MoveNext())
          acc[n++] = asMem((Resource)en.Value).toWorking();
        return new List(Sys.ResourceType, acc);
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Lifecycle
  //////////////////////////////////////////////////////////////////////////

    public override void refresh()
    {
      checkMountedWorking();
      lock (m_master)
      {
        this.m_obj = safe(m_master.m_obj);
        this.m_version = m_master.m_version;
      }
    }

    public override Resource create()
    {
      // get my parent's uri
      Uri myUri = uri();
      Uri parentUri = myUri.parent();
      if (parentUri == null)
        throw Err.make("Invalid Uri to mount `" + myUri + "`").val;

      // resolve my parent
      Resource parent = parentUri.resolve(null, Bool.False);
      if (parent == null)
        throw Err.make("Parent `" + parentUri + "` does not exist").val;

      // check parent is instance of MemResource
      if (!(parent is MemResource))
        throw Err.make("Parent `" + parentUri + "` not MemResource: " + parent.type()).val;

      asMem(parent).createChild(this);
      return this;
    }

    private void createChild(MemResource wc)
    {
      // legend:
      //   wc:  working child
      //   mc:  master child

      checkMountedWorking();

      if (wc.exists().val)
        throw Err.make("Already mounted `" + wc.uri() + "`").val;

      Str name = wc.uri().name();
      if (name == null || name.val.Length == 0)
        throw Err.make("Child must define a name: " + wc.type()).val;


      MemResource mc = wc.newResource(wc.uri());
      mc.m_isMaster = true;
      mc.m_obj = Resource.safe(wc.m_obj);

      lock (m_master)
      {
        if (m_master.m_children == null)
          m_master.m_children = new Hashtable(16);

        if (m_master.m_children[name] != null)
          throw Err.make("Duplicate name: " + name).val;

        m_master.m_children[name] = mc;

        wc.m_master = mc;
      }
    }

    public override void save()
    {
      checkMountedWorking();

      Obj safeObj = Resource.safe(m_obj);
      lock (m_master)
      {
        if (m_version != m_master.m_version)
          throw Err.make("Resource has been updated since last read `" + uri() + "`").val;
        m_master.m_obj = safeObj;
        m_version = ++m_master.m_version;
      }
    }

    public override void delete()
    {
      checkMountedWorking();

      Uri u = uri();
      Str name = u.name();
      Uri parentUri = u.parent();
      if (name == null || parentUri == null)
        throw UnsupportedErr.make("Cannot delete `" + u + "`").val;

      MemResource workingParent = asMem(parentUri.resolve());
      MemResource masterParent = workingParent.m_master;

      // remove from parent's master
      lock (masterParent)
      {
        masterParent.m_children.Remove(name);
      }

      // cleanup my master (shouldn't really matter)
      m_master.m_obj = null;
      m_master.m_version = -1;

      // cleanup myself
      this.m_master = null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Resolve
  //////////////////////////////////////////////////////////////////////////

    public override Resource resolve(Uri uri, Int startDepth)
    {
      List path = uri.m_path;
      int pathDepth = path.sz();
      MemResource p = this;

      for (int i=(int)startDepth.val; i<pathDepth; ++i)
      {
        lock (p)
        {
          if (p.m_children == null) return null;
          Resource r = (Resource)p.m_children[path.get(i)];
          if (r == null) return null;
          p = asMem(r);
        }
      }

      return p.toWorking();
    }

  //////////////////////////////////////////////////////////////////////////
  // Overrides
  //////////////////////////////////////////////////////////////////////////

    /**
     * Make a new instance of a Resource for this peer type.
     */
    public MemResource newResource(Uri uri)
    {
      return (MemResource)type().make(new List(Sys.ObjType, new Obj[] { uri, null }));
    }

    /**
     * Get a working instance of the peer which is safe to pass
     * to threads to work with the resource.  The working instance
     * should be insulated from further changes to this master
     * instance (until refreshed).
     */
    public Resource toWorking()
    {
      checkMaster();
      MemResource w = newResource(this.uri());
      w.m_master   = this;
      w.m_version  = this.m_version;
      w.m_needLoad = true;
      return w;
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    /**
     * Get the resource as a MemResource.
     */
    public static MemResource asMem(Resource r)
    {
      try
      {
        return (MemResource)r;
      }
      catch (System.InvalidCastException)
      {
        throw Err.make("Not a MemResource: " + r.type()).val;
      }
    }

    /**
     * Verify that this instance is an working instance.
     */
    public void checkWorking()
    {
      if (m_isMaster) throw UnsupportedErr.make("Not working instance").val;
    }

    /**
     * Verify that this instance is an unmounted working instance.
     */
    public void checkUnmountedWorking()
    {
      checkWorking();
      if (m_master != null)
        throw UnsupportedErr.make("Resource is already mounted `" + uri() + "`").val;
    }

    /**
     * Verify that this instance is an mounted working instance.
     */
    public void checkMountedWorking()
    {
      checkWorking();
      if (m_master == null)
        throw UnsupportedErr.make("Resource not mounted `" + uri() + "`").val;
    }

    /**
     * Verify that this instance is an mounted master instance.
     */
    public void checkMaster()
    {
      if (!m_isMaster)
        throw UnsupportedErr.make("Not master instance").val;
    }

    /**
     * Trap dynamic slot access.
     */
    public override Obj trap(Str name, List args)
    {
      string n = name.val;
      if (n == "dump") { dump(); return null; }
      return base.trap(name, args);
    }

    /**
     * Dump the state to standard out
     */
    public void dump()
    {
      System.Console.WriteLine(GetType().Name);
      System.Console.WriteLine("  isMaster = " + m_isMaster);
      System.Console.WriteLine("  uri      = " + uri());
      System.Console.WriteLine("  master   = " + m_master);
      System.Console.WriteLine("  needLoad = " + m_needLoad);
      System.Console.WriteLine("  obj      = " + m_obj);
      System.Console.WriteLine("  version  = " + m_version);
      System.Console.WriteLine("  children = " + m_children);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal bool m_isMaster;
    internal MemResource m_master;
    internal bool m_needLoad;
    internal Obj m_obj;
    internal int m_version;
    internal Hashtable m_children;

  }

}