//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Nov 07  Brian Frank  Creation
//

namespace Fan.Sys
{
  /// <summary>
  /// Resource models an object in the Uri namespace.  Resources
  /// delegate all methods to a ResourcePeer which implements the
  /// functionality in Java code.  Binding between a Resource and
  /// its ResourcePeer occur in the ResourcePeer constructor and
  /// last for the lifetime of both instances.
  /// </summary>
  public abstract class Resource : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static void make_(Resource self, Uri uri)
    {
      if (uri == null) throw ArgErr.make("Null uri").val;
      self.m_uri = uri;
    }

    public Resource(Uri uri) { make_(this, uri); }

    public Resource() {}

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Str toStr() { return m_uri.toStr(); }

    public override Type type() { return Sys.ResourceType; }

  //////////////////////////////////////////////////////////////////////////
  // Uri
  //////////////////////////////////////////////////////////////////////////

    public Uri uri() { return m_uri; }

    public Bool isDir() { return m_uri.isDir();   }

    public List path() { return m_uri.path(); }

    public Str pathStr() { return m_uri.pathStr(); }

    public Str name() { return m_uri.name(); }

    public Str basename() { return m_uri.basename(); }

    public Str ext() { return m_uri.ext(); }

  //////////////////////////////////////////////////////////////////////////
  // Resource Overrides
  //////////////////////////////////////////////////////////////////////////

    public virtual Bool exists()
    {
      throw UnsupportedErr.make(type() + ".exists").val;
    }

    public virtual Obj obj()
    {
      return null;
    }

    public virtual void obj(Obj obj)
    {
      throw UnsupportedErr.make(type() + ".obj").val;
    }

    public virtual List list()
    {
      return emptyList;
    }

    public virtual void refresh()
    {
      throw UnsupportedErr.make(type() + ".refresh").val;
    }

    public virtual Resource create()
    {
      throw UnsupportedErr.make(type() + ".create").val;
    }

    public virtual Resource add(Obj obj)
    {
      throw UnsupportedErr.make(type() + ".add").val;
    }

    public virtual void save()
    {
      throw UnsupportedErr.make(type() + ".save").val;
    }

    public virtual void delete()
    {
      throw UnsupportedErr.make(type() + ".delete").val;
    }

    public virtual Resource resolve(Uri uri, Int startDepth)
    {
      return null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    /**
     * Make a thread-safe copy of the specified object.
     * If it is immutable, then just return it; otherwise
     * we make a serialized copy.
     */
    public static Obj safe(Obj obj)
    {
      if (obj == null) return null;
      if (obj.isImmutable().val) return obj;
      Buf buf = Buf.make(256);
      buf.m_out.writeObj(obj);
      buf.flip();
      return buf.m_in.readObj();
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly List emptyList = new List(Sys.ResourceType).ro();

    internal Uri m_uri;

  }
}