//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Dec 07  Andy Frank  Creation
//

using DirectoryInfo = System.IO.DirectoryInfo;
using FileInfo = System.IO.FileInfo;
using FileSystemInfo = System.IO.FileSystemInfo;

namespace Fan.Sys
{
  /// <summary>
  /// RootResource is an internal type for managing the
  /// root of the resource namespace itself.
  /// </summary>
  public class RootResource : MemResource
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static RootResource make(Uri uri)
    {
      RootResource self = new RootResource();
      MemResource.make_(self, uri, null);
      return self;
    }

    public static RootResource makeMaster()
    {
      RootResource self = new RootResource();
      MemResource.make_(self, Uri.fromStr("/"), null);
      self.m_isMaster = true;
      return self;
    }

    public RootResource() {}

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Type type()
    {
      return Sys.RootResourceType;
    }

    public override Bool exists()
    {
      return Bool.True;
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Obj obj()
    {
      return null;
    }

    public override void obj(Obj obj)
    {
      throw UnsupportedErr.make("sys::RootResource.obj").val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Children
  //////////////////////////////////////////////////////////////////////////

    public override List list()
    {
      // MemResource super class mounted resources
      List list = base.list().rw();

      // `/sys`
      list.add(SysResource.root());

      // Sys.appDir
      FileSystemInfo[] files = new DirectoryInfo(Sys.AppDir).GetFileSystemInfos();
      for (int i=0; i<files.Length; ++i)
      {
        FileSystemInfo f = files[i];
        string uri = "/" + f.Name;
        if (f is DirectoryInfo) uri += "/";
        list.add(new LocalFile(Uri.fromStr(uri), f));
      }

      return list;
    }

    public override Resource resolve(Uri uri, Int startDepth)
    {
      // sanity check
      if (startDepth != Int.Zero)
        throw new System.Exception();

      // check if resolving myself
      if (uri.m_path.sz() == 0)
        return toWorking();

      // `/sys`
      if (uri.path(0) == "sys")
        return SysResource.root().resolve(uri, Int.One);

      // check MemResource super class for mounted resources
      Resource r = base.resolve(uri, startDepth);
      if (r != null) return r;

      // check Sys.appDir
      FileSystemInfo f = find(new DirectoryInfo(Sys.AppDir), uri.path(), 0);
      if (f != null && f.Exists)
      {
        if (f is DirectoryInfo && !uri.isDir().val)
          uri = Uri.fromStr(uri.pathStr().val + "/");
        return new LocalFile(uri, f);
      }

      // no joy
      return null;
    }

    FileSystemInfo find(DirectoryInfo dir, List path, int i)
    {
      if (i == path.sz()) return dir;  // we found it baby!

      string sub = (path.get(i) as Str).val;
      FileSystemInfo[] f = dir.GetFileSystemInfos(sub);

      if (f.Length == 0) return null;     // no match
      if (f[0] is FileInfo) return f[0];  // found it
      return find(f[0] as DirectoryInfo, path, ++i);  // need to keep looking
    }

  //////////////////////////////////////////////////////////////////////////
  // Unsupported
  //////////////////////////////////////////////////////////////////////////

    public override Resource create()
    {
      throw UnsupportedErr.make("sys::RootResource.create").val;
    }

    public override void save()
    {
      throw UnsupportedErr.make("sys::RootResource.save").val;
    }

    public override void delete()
    {
      throw UnsupportedErr.make("sys::RootResource.delete").val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

  }
}