//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   12 Mar 07  Andy Frank  Creation
//

using System;
using System.IO;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// SysInStream.
  /// </summary>
  public class SysInStream : InStream
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public static SysInStream make(Stream input, Int bufSize)
    {
      if (bufSize == null || bufSize.val == 0)
        return new SysInStream(input);
      else
        return new SysInStream(new BufferedStream(input, (int)bufSize.val));
    }

    public SysInStream(Stream input)
    {
      this.inStream = input;
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.SysInStreamType; }

  //////////////////////////////////////////////////////////////////////////
  // InStream
  //////////////////////////////////////////////////////////////////////////

    public override Int read() { int n = r(); return n < 0 ? null : Int.pos(n); }
    public override int r()
    {
      try
      {
        return inStream.ReadByte();
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public override Int readBuf(Buf buf, Int n)
    {
      try
      {
        int len = (int)n.val;
        buf.grow(buf.m_pos+len);
        int read = inStream.Read(buf.m_buf, buf.m_pos, len);
        if (read <= 0) return null;
        buf.m_pos  += read;
        buf.m_size += read;
        return Int.pos(read);
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public override InStream unread(Int n) { return unread((int)n.val); }
    public override InStream unread(int n)
    {
      try
      {
        // don't take the hit until we know we need to wrap
        // the raw input stream with a pushback stream
        if (!(inStream is PushbackStream))
          inStream = new PushbackStream(inStream, 1024);
        ((PushbackStream)inStream).Unread(n);
        return this;
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    /*
    public override Buf readAllBuf() { return m_in.readAllBuf(this); }

    public override Buf readBufFully(Buf buf, Int n) { return m_in.readBufFully(this, buf, n); }

    public override Int readU1() { return m_in.readU1(this); }

    public override Int readS1() { return m_in.readS1(this); }

    public override Int readU2() { return m_in.readU2(this); }

    public override Int readS2() { return m_in.readS2(this); }

    public override Int readU4() { return m_in.readU4(this); }

    public override Int readS4() { return m_in.readS4(this); }

    public override Int readS8() { return m_in.readS8(this); }

    public override Float readF4() { return m_in.readF4(this); }

    public override Float readF8() { return m_in.readF8(this); }

    public override Bool readBool() { return m_in.readBool(this); }

    public override Str readUtf() { return m_in.readUtf(this); }

    public override int rChar()
    {
      return m_charsetDecoder.decode(this);
    }

    public override Int readChar()
    {
      int ch = m_charsetDecoder.decode(this);
      return ch < 0 ? null : Int.pos(ch);
    }


//    public OutStream writeChars(Str s) { return writeChars(s.val, 0, s.val.length()); }
//    public OutStream writeChars(Str s, Int off) { return writeChars(s.val, (int)off.val, s.val.length()-(int)off.val); }
//    public OutStream writeChars(Str s, Int off, Int len) { return writeChars(s.val, (int)off.val, (int)len.val); }
//    public OutStream writeChars(String s, int off, int len)
//    {
//      try
//      {
//        writer().write(s, off, len);
//        return this;
//      }
//      catch (java.io.IOException e)
///      {
//        throw IOErr.make(e);
//      }
//    }


    public override InStream unreadChar(Int c)
    {
      m_charsetEncoder.encode((char)c.val, this);
      return this;
    }

    public override Str readLine() { return m_in.readLine(this, Int.m_pos[4096]); }
    public override Str readLine(Int max) { return m_in.readLine(this, max); }

    public override List readAllLines() { return m_in.readAllLines(this); }

    public override void eachLine(Func f) { m_in.eachLine(this, f); }

    public override Str readAllStr() { return m_in.readAllStr(this, Bool.True); }
    public override Str readAllStr(Bool normalizeNewlines)  { return m_in.readAllStr(this, normalizeNewlines); }

    public override Map readProps() { return m_in.readProps(this); }

    public override Charset charset()
    {
      return m_charset;
    }

    public override void charset(Charset charset)
    {
      m_charsetDecoder = charset.newDecoder();
      m_charsetEncoder = charset.newEncoder();
      m_charset = charset;
    }

    public override Int pipe(OutStream ot) { return m_in.pipe(this, ot, null, Bool.True); }
    public override Int pipe(OutStream ot, Int n) { return m_in.pipe(this, ot, n, Bool.True); }
    public override Int pipe(OutStream ot, Int n, Bool close) { return m_in.pipe(this, ot, n, close); }

    public override Bool close()
    {
      try
      {
        m_in.Close();
        return Bool.True;
      }
      catch (IOException)
      {
        return Bool.False;
      }
    }
    */

    public override Bool close()
    {
      try
      {
        if (inStream != null) inStream.Close();
        return Bool.True;
      }
      catch (IOException)
      {
        return Bool.False;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // PushbackStream
  //////////////////////////////////////////////////////////////////////////

    internal class PushbackStream : Stream
    {
      public PushbackStream(Stream parent, int size)
      {
        this.parent = parent;
        this.pushback = new byte[size];
      }

      //
      // Properties
      //

      public override bool CanRead  { get { return parent.CanRead;  } }
      public override bool CanWrite { get { return parent.CanWrite; } }
      public override bool CanSeek  { get { return false; } } // don't allow seeking
      public override long Length   { get { return parent.Length; } }
      public override long Position
      {
        get { return parent.Position;  }
        set { parent.Position = value; }
      }

      //
      // Methods
      //

      public void Unread(int b)
      {
        if (pos >= pushback.Length)
          throw new IOException("Not enough room in pushback buffer");
        pushback[pos++] = (byte)b;
      }

      public override int Read(byte[] buf, int offset, int count)
      {
        // check pushback first
        if (pos > 0)
        {
          int len = Math.Min(pos, count);
          Array.Copy(pushback, pos-len, buf, offset, len);
          pos -= len;
          offset += len;
          count -= len;

          // short-circuit if buf filled with pushback
          if (count == 0) return len;
        }

        // grab remaining bytes
        return parent.Read(buf, offset, count);
      }

      public override void Write(byte[] buf, int offset, int count)
      {
        // This is purely an input stream, so Write not allowed
        throw new IOException("PushbackStream does not allow Write()");
      }

      public override long Seek(long offset, SeekOrigin origin)
      {
        // We don't allow seeking
        throw new IOException("PushbackStream does not allow Seek()");
      }

      public override void SetLength(long val)
      {
        // We don't allow seeking
        throw new IOException("PushbackStream does not allow SetLength()");
      }

      public override void Close() { parent.Close(); }
      public override void Flush() { parent.Flush(); }

      //
      // Fields
      //

      private Stream parent;
      private byte[] pushback;
      private int pos = 0;
    }

  //////////////////////////////////////////////////////////////////////////
  // .NET Conversion
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Get a System.IO.Stream for the specified input stream.
    /// </summary>
    public static Stream net(InStream ins)
    {
      if (ins is SysInStream)
        return ((SysInStream)ins).inStream;
      else
        return new NetInputStream (ins);
    }

    internal class NetInputStream : Stream
    {
      public NetInputStream(InStream ins) { this.ins = ins; }

      // Properties
      public override bool CanRead  { get { return true;  } }
      public override bool CanWrite { get { return false; } }
      public override bool CanSeek  { get { return false; } }
      public override long Length   { get { return -1; } }
      public override long Position { get { return -1; } set {} }

      // Methods
      public override int Read(byte[] b, int off, int len)
      {
        buf.m_buf = b;
        buf.m_pos = off;
        buf.m_size = b.Length;
        Int n = ins.readBuf(buf, Int.make(len));
        if (n == null) return -1;
        return (int)n.val;
      }
      public override long Seek(long off, SeekOrigin origin) { return -1; }
      public override void SetLength(long val) {}
      public override void Flush() {}
      public override void Write(byte[] b, int off, int len) {}
      public override void Close() { ins.close(); }

      // Fields
      private InStream ins;
      private Buf buf = Buf.make();
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal Stream inStream;

  }
}