//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Dec 07  Andy Frank  Creation
//

using System.Collections;

namespace Fan.Sys
{
  /// <summary>
  /// SysResource is used to wrap built-in immutable system resources.
  /// </summary>
  public class SysResource : Resource
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static SysResource root()
    {
      if (m_root == null)
      {
        m_root = make("/sys/", new NavDir())
          .add(make("/sys/logs/",    new LogsDir()))
          .add(make("/sys/threads/", new ThreadsDir()));
      }
      return m_root;
    }

    public static SysResource make(string uri, Dir dir)
    {
      SysResource self = new SysResource();
      self.m_uri = Uri.fromStr(uri);
      self.m_dir = dir;
      return self;
    }

    public static SysResource make(string uri, Obj obj)
    {
      if (!obj.isImmutable().val)
        throw NotImmutableErr.make(obj.type().toStr()).val;

      SysResource self = new SysResource();
      self.m_uri = Uri.fromStr(uri);
      self.m_dir = new EmptyDir();
      self.m_obj = obj;
      return self;
    }

    private SysResource() {}

  //////////////////////////////////////////////////////////////////////////
  // Resource
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.SysResourceType; }

    public override Bool exists() { return Bool.True; }

    public override Obj obj() { return m_obj; }

    public override List list() { return m_dir.list(); }

    public override Resource resolve(Uri uri, Int startDepth)
    {
      if (uri.m_path.size() == startDepth) return this;
      SysResource kid = m_dir.resolve((Str)uri.path().get(startDepth));
      if (kid == null) return null;
      return kid.resolve(uri, startDepth.plus(Int.One));
    }

  //////////////////////////////////////////////////////////////////////////
  // Dir
  //////////////////////////////////////////////////////////////////////////

    public abstract class Dir
    {
      public abstract List list();
      public abstract SysResource resolve(Str name);
    }

  //////////////////////////////////////////////////////////////////////////
  // EmptyDir
  //////////////////////////////////////////////////////////////////////////

    internal class EmptyDir : Dir
    {
      public override List list() { return Resource.emptyList; }
      public override SysResource resolve(Str name) { return null; }
    }

  //////////////////////////////////////////////////////////////////////////
  // NavDir
  //////////////////////////////////////////////////////////////////////////

    SysResource add(SysResource kid)
    {
      ((NavDir)m_dir).add(kid);
      return this;
    }

    internal class NavDir : Dir
    {
      public override List list()
      {
        return m_list.ro();
      }

      public override SysResource resolve(Str name)
      {
        return (SysResource)m_map[name];
      }

      public void add(SysResource kid)
      {
        m_map[kid.name()] = kid;
        m_list.add(kid);
      }

      Hashtable m_map = new Hashtable();
      List m_list = new List(Sys.ResourceType);
    }

  //////////////////////////////////////////////////////////////////////////
  // ThreadsDir
  //////////////////////////////////////////////////////////////////////////

    internal class ThreadsDir : Dir
    {
      public override List list()
      {
        List threads = Thread.list();
        List acc = new List(Sys.ResourceType, threads.sz());
        for (int i=0; i<threads.sz(); ++i)
        {
          Thread t = (Thread)threads.get(i);
          acc.add(make("/sys/threads/" + t.name(), t));
        }
        return acc;
      }

      public override SysResource resolve(Str name)
      {
        Thread t = Thread.find(name, Bool.False);
        if (t == null) return null;
        return make("/sys/threads/" + t.name(), t);
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // LogsDir
  //////////////////////////////////////////////////////////////////////////

    internal class LogsDir : Dir
    {
      public override List list()
      {
        List logs = Log.list();
        List acc = new List(Sys.ResourceType, logs.sz());
        for (int i=0; i<logs.sz(); ++i)
        {
          Log log = (Log)logs.get(i);
          acc.add(make("/sys/logs/" + log.name(), log));
        }
        return acc;
      }

      public override SysResource resolve(Str name)
      {
        Log log  = Log.find(name, Bool.False);
        if (log == null) return null;
        return make("/sys/logs/" + log.name(), log);
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private static SysResource m_root;

    Obj m_obj;
    Dir m_dir;
  }
}