//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   23 Feb 07  Andy Frank  Creation
//

using System;
using System.Collections;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using NThread = System.Threading.Thread;
using ThreadPool = System.Threading.ThreadPool;
using ThreadInterruptedException = System.Threading.ThreadInterruptedException;

namespace Fan.Sys
{
  /// <summary>
  /// Thread models a thread of execution within a process.
  /// </summary>
  public class Thread : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Thread make() { return make(null, null); }
    public static Thread make(Str name) { return make(name, null); }
    public static Thread make(Str name, Func run)
    {
      Thread t = new Thread();
      make_(t, name, run);
      return t;
    }

    public static void make_(Thread t) { make_(t, null, null); }
    public static void make_(Thread t, Str name) { make_(t, name, null); }
    public static void make_(Thread t, Str name, Func run)
    {
      lock (topLock)
      {
        // check run method
        if (run != null && !run.isImmutable().val)
          throw NotImmutableErr.make("Run method not const: " + run).val;

        // auto generate name if null
        if (name == null)
          name = Str.make(t.type().pod().name() + "." + t.type().name() + "." + (autoNameCount++));

        // verify name is valid
        Uri.checkName(name);

        // verify unique
        if (byName[name] != null)
          throw ArgErr.make("Duplicate thread name: " + name).val;

        // init and put into map
        t.m_name  = name;
        t.m_run   = run;
        t.m_state = NEW;
        byName[name] = t;
      }
    }

    public Thread(string name)
    {
      this.m_name  = Str.make(name);
      this.m_state = NEW;
    }

    public Thread(Str name)
    {
      lock (topLock)
      {
        // verify unique
        if (byName[name] != null)
          throw ArgErr.make("Duplicate thread name: " + name).val;

        // put into map
        byName[name] = this;
      }

      this.m_name  = name;
      this.m_state = NEW;
    }

    public Thread()
    {
    }

  //////////////////////////////////////////////////////////////////////////
  // Management
  //////////////////////////////////////////////////////////////////////////

    public static Thread find(Str name) { return find(name, Bool.True); }
    public static Thread find(Str name, Bool check)
    {
      lock (topLock)
      {
        Thread thread = (Thread)byName[name];
        if (thread != null) return thread;
        if (check.val) throw UnknownThreadErr.make(name).val;
        return null;
      }
    }

    public static List list()
    {
      lock (topLock)
      {
        Thread[] arr = new Thread[byName.Count];
        byName.Values.CopyTo(arr, 0);
        return new List(Sys.ThreadType, arr).ro();
      }
    }

    public static Thread current()
    {
      if (m_threadLookup == null)
        throw Err.make("Current thread not a Fan thread: " + NThread.CurrentThread).val;
      return m_threadLookup;
    }

    public static Map locals()
    {
      if (m_locals == null) m_locals = new Map(Sys.StrType, Sys.ObjType);
      return m_locals;
    }

    [ThreadStatic] static Map m_locals;
    [ThreadStatic] static Thread m_threadLookup;

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Bool equals(Obj obj)
    {
      return this == obj ? Bool.True : Bool.False;
    }

    public override int GetHashCode()
    {
      return m_name.GetHashCode();
    }

    public override Int hash()
    {
      return m_name.hash();
    }

    public override Str toStr()
    {
      return m_name;
    }

    public override Type type()
    {
      return Sys.ThreadType;
    }

    public Str name()
    {
      return m_name;
    }

    public void trace() { trace(Sys.@out()); }
    public void trace(OutStream @out)
    {
      if (m_thread == null) return;

      StackTrace st = new StackTrace(true);
      for(int i=1; i<st.FrameCount; i++)
      {
        StackFrame sf = st.GetFrame(i);
        MethodBase mb = sf.GetMethod();
        string loc = sf.GetFileName();

        // remove the filepath if it exists
        if (loc == null) loc = "Unknown Source";
        else
        {
          int index = loc.LastIndexOf("\\");
          if (index != -1) loc = loc.Substring(index+1);
          loc += ":" + sf.GetFileLineNumber();
        }

        StringBuilder sb = new StringBuilder();
        sb.Append(mb.ReflectedType).Append(".").Append(mb.Name);
        sb.Append("(").Append(loc).Append(")");
        @out.printLine(Str.make(sb.ToString()));
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // State
  //////////////////////////////////////////////////////////////////////////

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Bool isNew()
    {
      return Bool.make(m_state == NEW);
    }

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Bool isRunning()
    {
      return Bool.make(m_state == RUNNING);
    }

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Bool isDead()
    {
      return Bool.make(m_state == DEAD);
    }

  //////////////////////////////////////////////////////////////////////////
  // Lifecycle
  //////////////////////////////////////////////////////////////////////////

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Thread start()
    {
      if (m_state != NEW) throw Err.make("Invalid state for Thread.start").val;
      m_state = RUNNING;
      ThreadPool.QueueUserWorkItem(doRun, this);
      return this;
    }

    private static void doRun(object state)
    {
      // run attached thread
      Thread attached = state as Thread;
//System.Console.WriteLine(" -- doRun \t\"" + attached.m_name + "\" [" + NThread.CurrentThread.ManagedThreadId + "]");
      attached.m_thread = NThread.CurrentThread;
      m_threadLookup = attached;
      try
      {
        Obj result = attached.run();
        lock (attached)
        {
          attached.m_runResult = result;
        }
      }
      catch (Exception e)
      {
        Err.dumpStack(e);
      }

      // attached thread is now dead
      m_threadLookup = null;
      attached.m_thread = null;
      attached.stop();
    }

    [MethodImpl(MethodImplOptions.Synchronized)]
    public void stop()
    {
      if (m_state == DEAD) return;
      m_state = DEAD;
      lock (topLock) { byName.Remove(m_name); }
      stopMessages();
      Monitor.PulseAll(this);
      if (m_thread != null) m_thread.Interrupt();
//System.Console.WriteLine(" -- stopped \t\"" + m_name + "\" [" + NThread.CurrentThread.ManagedThreadId + "] " + (m_thread == null));
    }

    public Obj join() { return join(null); }

    [MethodImpl(MethodImplOptions.Synchronized)]
    public Obj join(Duration duration)
    {
      try
      {
        if (m_state == NEW) throw Err.make("Thread not started yet").val;

        if (m_state == RUNNING)
        {
          if (duration != null)
            Monitor.Wait(this, (int)duration.millis());
          else
            while (m_state == RUNNING) Monitor.Wait(this);
        }

        Obj r = m_runResult;
        m_runResult = null;
        return r;
      }
      catch (ThreadInterruptedException e)
      {
        throw InterruptedErr.make(e).val;
      }
    }

    public static void sleep(Duration duration)
    {
      try
      {
        NThread.Sleep((int)duration.millis());
      }
      catch (ThreadInterruptedException e)
      {
        throw InterruptedErr.make(e).val;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Runner
  //////////////////////////////////////////////////////////////////////////

    static void runner(object state)
    {
      // TODO
    }

  //////////////////////////////////////////////////////////////////////////
  // Run
  //////////////////////////////////////////////////////////////////////////

    public virtual Obj run()
    {
      if (m_run != null)
      {
        return m_run.call1(this);
      }
      else
      {
        System.Console.WriteLine("WARNING: Thread.run not overridden");
        return null;
      }
    }

    public void loop(Func received)
    {
      // ensure only called by myself
      if (current() != this)
        throw Err.make("Thread.current != this").val;

      // ensure received not null
      if (received == null)
        throw NullErr.make("received callback null").val;

      // main loop
      while (m_state == RUNNING)
      {
        try
        {
          dispatch(received, dequeue());
        }
        catch (Exception e)
        {
          if (m_state == RUNNING) Err.dumpStack(e);
        }
      }
    }

    private void dispatch(Func received, Message msg)
    {
      try
      {
        Obj result = received.call1(msg.obj);
        result = Resource.safe(result);
        msg.finish(MSG_FINISH_OK, result);
      }
      catch (Err.Val e)
      {
        msg.finish(MSG_FINISH_ERR, e.m_err);
      }
      catch (Exception e)
      {
        msg.finish(MSG_FINISH_ERR, Err.make(e));
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Messaging
  //////////////////////////////////////////////////////////////////////////

    public Obj sendSync(Obj obj)
    {
      obj = Resource.safe(obj);

      try
      {
        Message msg = new Message(MSG_SYNC, obj);
        enqueue(msg);
        return msg.waitUntilFinished();
      }
      catch(ThreadInterruptedException e)
      {
        throw InterruptedErr.make(e).val;
      }
    }

    public void sendAsync(Obj obj)
    {
      obj = Resource.safe(obj);

      try
      {
        enqueue(new Message(MSG_ASYNC, obj));
      }
      catch(ThreadInterruptedException e)
      {
        throw InterruptedErr.make(e).val;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Queue
  //////////////////////////////////////////////////////////////////////////

    [MethodImpl(MethodImplOptions.Synchronized)]
    void stopMessages()
    {
      for (Message m = m_head; m != null; m = m.next)
        m.finish(MSG_STOPPED, null);
      m_head = m_tail = null;
      m_size = 0;
    }

    [MethodImpl(MethodImplOptions.Synchronized)]
    Message dequeue()
    {
      // wait until we have a message
      while (m_state == RUNNING && m_size == 0)
        Monitor.Wait(this);

      // read message of head of linked list
      Message m = m_head;
      m_head = m.next;
      if (m_head == null) m_tail = null;
      m.next = null;
      m_size--;

      // return message
      return m;
    }

    [MethodImpl(MethodImplOptions.Synchronized)]
    void enqueue(Message m)
    {
//System.Console.WriteLine(" -- enqueue \t\"" + m_name + "\" [" + NThread.CurrentThread.ManagedThreadId + "] " + m_state);

      // ensure new or running
      if (m_state > RUNNING)
        throw Err.make("thread not active").val;

      // flow control wait if at max
      if (m_size >= maxQueueSize) Monitor.Wait(this);

      // add to tail of linked list
      if (m_tail == null) { m_head = m_tail = m; m.next = null; }
      else { m_tail.next = m; m_tail = m; }
      m_size++;
      if (m_size > m_peek) m_peek = m_size;

      // notify get thread
      Monitor.Pulse(this);
    }

  //////////////////////////////////////////////////////////////////////////
  // Message
  //////////////////////////////////////////////////////////////////////////

    internal class Message
    {
      public Message(int state, Obj obj)
      {
        this.state = state;
        this.obj = obj;
      }

      [MethodImpl(MethodImplOptions.Synchronized)]
      internal Obj waitUntilFinished()
      {
        if (state == MSG_SYNC) Monitor.Wait(this);

        switch (state)
        {
          case MSG_SYNC:       throw InterruptedErr.make("sendSync timed out").val;
          case MSG_FINISH_OK:  return obj;
          case MSG_FINISH_ERR: throw ((Err)obj).val;
          case MSG_STOPPED:    throw InterruptedErr.make("thread stopped").val;
          default: throw new Exception(""+state);
        }
      }

      internal void finish(int newState, Obj obj)
      {
        if (this.state == MSG_ASYNC)
        {
          if (newState == MSG_FINISH_ERR)
            ((Err)obj).trace();
        }
        else
        {
          lock (this)
          {
            this.state = newState;
            this.obj = obj;
            Monitor.Pulse(this);
          }
        }
      }

      internal int state;       // sync/async in, ok/err out
      internal Obj obj;         // message in, return/err out
      internal Message next;    // queue linked list
    }

    const int MSG_ASYNC      = 0;
    const int MSG_SYNC       = 1;
    const int MSG_FINISH_OK  = 2;
    const int MSG_FINISH_ERR = 3;
    const int MSG_STOPPED    = 4;

  //////////////////////////////////////////////////////////////////////////
  // State Constants
  //////////////////////////////////////////////////////////////////////////

    const int NEW     = 0;
    const int RUNNING = 1;
    const int DEAD    = 2;

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private static Object topLock = new Object();   // top level lock
    private static Hashtable byName = new Hashtable();  // String -> Thread
    private static int autoNameCount = 0;           // auto-generate unique name
    private static int maxQueueSize = 1000;         // max messages to queue

    private Str m_name;                 // thread name
    private int m_state;                // current state
    private NThread m_thread;           // .NET thread if attached
    private Message m_head, m_tail;     // message queue linked list
    private int m_size, m_peek;         // message queue size
    private Func m_run;                 // run method
    private Obj m_runResult;            // return of run method

  }
}