//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 07  Andy Frank  Creation
//

using Fan.Sys;

namespace Fanx.Tools
{
  /// <summary>
  /// Fan Daemon
  /// </summary>
  ///
  public class Fand : Tool
  {

  /////////////////////////////////////////////////////////////////////////
  // Run
  //////////////////////////////////////////////////////////////////////////

    public static int run(string reserved)
    {
      MainThread t = new MainThread();
      t.start().join();
      return t.ret;
    }

    class MainThread : Thread
    {
      public MainThread() : base(Str.make("main")) {}
      public override Obj run()
      {
        ret = doRun();
        return null;
      }
      public int ret;
    }

    static int doRun()
    {
      try
      {
        string[] args = System.Environment.GetCommandLineArgs();
        if (!processArgs(args)) return -1;
        if (!init()) return -1;
        if (!loadServices()) return -1;
        return waitToExit();
      }
      catch (System.Exception e)
      {
        Err.dumpStack(e);
        return -1;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Process Arguments
  //////////////////////////////////////////////////////////////////////////

    static bool processArgs(string[] args)
    {
      if (args.Length == 1) { help(); return false; }

      string appName = null;
      for (int i=1; i<args.Length; ++i)
      {
        string a = System.String.Intern(args[i]);
        if (a.Length == 0) continue;
        if (a == "-help" || a == "-h" || a == "-?")
        {
          help();
          return false;
        }
        else if (a == "-version")
        {
          Fan.version("Fan Daemon");
          return false;
        }
        else if (a[0] == '-')
        {
          System.Console.WriteLine("WARNING: Unknown option " + a);
        }
        else
        {
          if (appName == null)
            appName = a;
          else
            println("WARNING: Argument ignored '" + a + "'");
        }
      }

      if (appName == null)
      {
        help();
        return false;
      }

      // system property for Sys.appDir
      Sys.putProperty("fan.appDir", "$home/fand/" + appName + "/");
      return true;
    }

    static void help()
    {
      println("Usage:");
      println("  fand [options] <appName>");
      println("Options:");
      println("  -help, -h, -?  print usage help");
      println("  -version       print version information");
    }

    static void println(string s)
    {
      System.Console.WriteLine(s);
    }

  //////////////////////////////////////////////////////////////////////////
  // Init
  //////////////////////////////////////////////////////////////////////////

    /**
     * Initialize all the directories and internal state.
     */
    static bool init()
    {
      // resolve app directory
      appDir = Sys.appDir();
      if (!appDir.exists().val)
      {
        log.error("File not found `" + appDir + "`");
        return false;
      }

      log.info("Fand version " + Sys.SysPod.version());
      log.info("App dir `" + appDir + "`");
      return true;
    }

  //////////////////////////////////////////////////////////////////////////
  // Load Services
  //////////////////////////////////////////////////////////////////////////

    /**
     * Find all the "appDir/services/*.fog" files and
     * load them as services.
     */
    static bool loadServices()
    {
      List services = new List(Sys.ObjType);

      // load all the fog files under /appDir/services
      File dir = appDir.plus(Uri.fromStr("services/"));
      List files = dir.list();
      for (int i=0; i<files.sz(); ++i)
      {
        Thread t = loadService((File)files.get(i));
        if (t != null) services.add(t);
      }

      // if no services defined, bail
      if (services.sz() == 0)
      {
        log.error("No services defined under `" + dir + "`");
        return false;
      }

      // now start them all
      for (int i=0; i<services.sz(); ++i)
      {
        Thread t = (Thread)services.get(i);
        t.start();
      }

      return true;
    }

    /**
     * Given a file attempt to deserialize it as a Thread.
     */
    static Thread loadService(File f)
    {
      try
      {
        if (f.ext() == null) return null;
        if (f.ext().val != "fog") return null;

        Map options = new Map(Sys.StrType, Sys.ObjType);
        options.add(Str.make("makeArgs"), new List(Sys.ObjType, new Obj[] { f.basename() }));
        Obj obj = f.readObj(options);

        if (!(obj is Thread))
        {
          log.error("Cannot load `" + f + "`; Not a sys::Thread: " + obj.type());
          return null;
        }

        return (Thread)obj;
      }
      catch (System.Exception e)
      {
        log.error("Cannot load `" + f + "`", e);
        return null;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Wait To Exit
  //////////////////////////////////////////////////////////////////////////

    static int waitToExit()
    {
      isAlive = true;
      while (isAlive)
      {
        try
        {
          System.Threading.Thread.Sleep(60000);
        }
        catch (System.Exception e)
        {
          log.error("waitToExit", e);
        }
      }
      return 0;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    static Log log = Log.get("fand");
    static File appDir;
    static bool isAlive;

  }
}
