//
// Copyright (c) 2008, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   01 Jan 08  John Sublett  Creation
//

**
** TypeGraph is used to sort a set of types based on
** their dependencies on each other.  A set of types
** cannot be sorted if the types have cyclic dependencies.
**
class TypeGraph
{
  **
  ** Add a type to the graph without specifying a dependency.
  **
  Void addType(Type t)
  {
    node(t)
  }

  **
  ** Add a dependency from the 'from' type to the 'to' type.
  **
  Void addDependency(Type from, Type to)
  {
    parent := node(to)
    child := node(from)

    parent.addChild(child);
  }

  **
  ** Get the list of types in the graph sorted by dependency.
  ** A type will always be later in the list than the types
  ** it depends on.  That is, if type B depends on type A, the
  ** sorted result will be '[A, B]'.
  **
  Type[] sort()
  {
    return rsort.reverse
  }

  **
  ** Get the list of types in the graph sorted in reverse order.
  **
  Type[] rsort()
  {
    nodes.each |TypeGraphNode n| { n.nodeState = NodeState.unvisited }

    Type[] rsorted := Type[,]
    nodes.each |TypeGraphNode n| { n.visit(rsorted) }
    return rsorted
  }

  **
  ** Get the existing node for the specified
  ** type or create a new node if one does not
  ** already exist.
  **
  private TypeGraphNode node(Type nodeType)
  {
    n := nodes[nodeType.qname]
    if (n == null)
    {
      n = TypeGraphNode.make(nodeType)
      nodes[nodeType.qname] = n
    }
    return n
  }

  **
  ** The list of nodes in the graph.  A node wraps
  ** a Type and contains a child node for each dependent
  ** Type.
  **
  private Str:TypeGraphNode nodes := Str:TypeGraphNode[:]
}

//////////////////////////////////////////////////////////////////////////
// TypeGraphNode
//////////////////////////////////////////////////////////////////////////

internal class TypeGraphNode
{
  new make(Type nodeType)
  {
    this.nodeType = nodeType
  }

  Void visit(Type[] rsorted, TypeGraphNode[] path := TypeGraphNode[,])
  {
    path.add(this)
    switch (nodeState)
    {
      case NodeState.unvisited:
        nodeState = NodeState.visiting
        if (!children.isEmpty)
          children.each |TypeGraphNode child| { child.visit(rsorted, path) }
        nodeState = NodeState.visited
        rsorted.add(nodeType)

      case NodeState.visiting:
        pathStr := path.join(" > ") |TypeGraphNode n->Str| { return n.nodeType.toStr }
        throw Err.make("Cycle detected: $pathStr")

      case NodeState.visited:
        return
    }
  }

  Void addChild(TypeGraphNode child)
  {
    children.add(child)
  }

  Type nodeType
  TypeGraphNode[] children := TypeGraphNode[,]
  NodeState nodeState := NodeState.unvisited
}

//////////////////////////////////////////////////////////////////////////
// NodeState
//////////////////////////////////////////////////////////////////////////

internal enum NodeState { visited, unvisited, visiting }