//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   30 Jun 07  Brian Frank  Creation
//
package fan.sql;

import java.sql.*;
import java.util.StringTokenizer;
import fan.sys.*;
import fan.sql.Connection;

public class ConnectionPeer
{

//////////////////////////////////////////////////////////////////////////
// Peer Factory
//////////////////////////////////////////////////////////////////////////

  public static ConnectionPeer make(Connection fan)
  {
    return new ConnectionPeer();
  }

//////////////////////////////////////////////////////////////////////////
// Lifecycle
//////////////////////////////////////////////////////////////////////////

  public static Connection open(Str database, Str username, Str password)
  {
    try
    {
      Connection self = Connection.make();
      self.peer.jconn = DriverManager.getConnection(database.val, username.val, password.val);
      return self;
    }
    catch (SQLException e)
    {
      throw err(e);
    }
  }

  public Bool isClosed(Connection self)
  {
    try
    {
      return Bool.make(jconn.isClosed());
    }
    catch (SQLException e)
    {
      throw err(e);
    }
  }

  public Bool close(Connection self)
  {
    try
    {
      jconn.close();
      return Bool.True;
    }
    catch (Throwable e)
    {
      e.printStackTrace();
      return Bool.False;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Database metadata
//////////////////////////////////////////////////////////////////////////

  public List tables(Connection self)
  {
    try
    {
      DatabaseMetaData dbData = jconn.getMetaData();
      ResultSet tables =
        dbData.getTables(null,  // catalog
                         null,  // schema pattern
                         null,  // table name pattern
                         null); // types
    
      int nameIndex = tables.findColumn("TABLE_NAME");
      List tableList = new List(Sys.StrType, 32);
      while (tables.next())
      {
        String tableName = tables.getString(nameIndex);
        tableList.add(Str.make(tableName));
      }
      tables.close();
    
      return tableList.ro();
    }
    catch(SQLException ex)
    {
      throw err(ex);
    }
  }
  
  public Obj tableRow(Connection self, Str tableName)
  {
    try
    {
      DatabaseMetaData dbData = jconn.getMetaData();
      ResultSet columns = dbData.getColumns(null, null, tableName.val, null);
  
      // map the meta-data to a dynamic type
      Type t = Type.makeDynamic(listOfRow);
      
      int nameIndex = columns.findColumn("COLUMN_NAME");
      int typeIndex = columns.findColumn("DATA_TYPE");
      int typeNameIndex = columns.findColumn("TYPE_NAME");
      int colIndex = 0;
      while (columns.next())
      {
        String name = columns.getString(nameIndex);
        String typeName = columns.getString(typeNameIndex);
        Type fanType = SqlUtil.sqlToFanType(columns.getInt(typeIndex));
        if (fanType == null)
        {
          System.out.println("WARNING: Cannot map " + typeName + " to Fan type");
          fanType = Sys.StrType;
        }
        t.add(Col.make(Int.pos(colIndex++), Str.make(name), fanType, Str.make(typeName), null));
      }
      
      if (colIndex == 0)
        throw SqlErr.make(Str.make("Table not found: " + tableName)).val;
      
      Row row = (Row)t.make();
      row.peer.cells = new Obj[t.fields().sz()];;
      return row;
    }
    catch(SQLException ex)
    {
      throw err(ex);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Query
//////////////////////////////////////////////////////////////////////////

  public List query(Connection self, Str sql, Type of)
  {
    java.sql.Statement stmt = null;
    try
    {
      // execute the query
      stmt = jconn.createStatement();
      ResultSet rs = stmt.executeQuery(sql.val);
      return toRows(rs, of);
    }
    catch (SQLException e)
    {
      throw err(e);
    }
    finally
    {
      try { if (stmt != null) stmt.close(); } catch (Exception e) {}
    }
  }

  /**
   * Convert the result set to a list of the 'of' type.
   */
  List toRows(ResultSet rs, Type of)
    throws SQLException
  {
    List rows = null;
    if (of == null)
    {
      Type dtype = makeDynamicType(rs);
      rows = new List(dtype);
      while (rs.next())
        rows.add(makeDynamicRow(rs, dtype));
    }
    else
    {
      Field[] fields = mapTypeFields(rs, of);
      rows = new List(of);
      while (rs.next())
        rows.add(makeTypedRow(rs, of, fields));
    }

    return rows;
  }
  
  public void queryEach(Connection self, Str sql, Type of, Func eachFunc)
  {
    java.sql.Statement stmt = null;
    try
    {
      // execute the query
      stmt = jconn.createStatement();
      ResultSet rs = stmt.executeQuery(sql.val);
      each(rs, of, eachFunc);
    }
    catch (SQLException e)
    {
      throw err(e);
    }
    finally
    {
      try { if (stmt != null) stmt.close(); } catch (Exception e) {}
    }
  }

  /**
   * Invoke the 'eachFunc' on every row in the result.
   */
  void each(ResultSet rs, Type of, Func eachFunc)
    throws SQLException
  {
    if (of == null)
    {
      Type dtype = makeDynamicType(rs);
      while (rs.next())
        eachFunc.call1(makeDynamicRow(rs, dtype));
    }
    else
    {
      Field[] fields = mapTypeFields(rs, of);
      while (rs.next())
        eachFunc.call1(makeTypedRow(rs, of, fields));
    }
  }

  /**
   * Make a dynamic type to map the columns of the
   * result set.
   */
  Type makeDynamicType(ResultSet rs)
    throws SQLException
  {
    // map the meta-data to a dynamic type
    Type t = Type.makeDynamic(listOfRow);
    ResultSetMetaData meta = rs.getMetaData();
    int numCols = meta.getColumnCount();
    for (int i=0; i<numCols; ++i)
    {
      String name = meta.getColumnName(i+1);
      String typeName = meta.getColumnTypeName(i+1);
      Type fanType = SqlUtil.sqlToFanType(meta.getColumnType(i+1));
      if (fanType == null)
      {
        System.out.println("WARNING: Cannot map " + typeName + " to Fan type");
        fanType = Sys.StrType;
      }
      t.add(Col.make(Int.pos(i), Str.make(name), fanType, Str.make(typeName), null));
    }
    
    return t;
  }

  /**
   * Make a row of the specified dynamic type
   */
  Obj makeDynamicRow(ResultSet rs, Type of)
    throws SQLException
  {
    if (!of.isDynamic().val)
      throw SqlErr.make("Expecting dynamic type, not " + of).val;

    Row row = (Row)of.make();
    int numCols = rs.getMetaData().getColumnCount();
    Obj[] cells = new Obj[numCols];
    row.peer.cells = cells;
    for (int i=0; i<numCols; ++i)
      cells[i] = SqlUtil.toObj(rs, i+1);
    return row;
  }

  /**
   * For a static type, map the columns in the result
   * set to fields on the 'of' type.
   */
  Field[] mapTypeFields(ResultSet rs, Type of)
    throws SQLException
  {
    // map the meta-data to Fields
    ResultSetMetaData meta = rs.getMetaData();
    int numCols = meta.getColumnCount();
    Field[] fields = new Field[numCols];
    for (int i=0; i<numCols; ++i)
    {
      String name = meta.getColumnName(i+1);
      Type fanType = SqlUtil.sqlToFanType(meta.getColumnType(i+1));
      Field f = of.field(name, false);
      if (f == null)
        throw SqlErr.make(Str.make("Col does not map to field: " + name)).val;
      if (!fanType.fits(f.of()).val)
        throw SqlErr.make(Str.make("Col type does not match field: " + name + ": " + f.of() + " != " + fanType)).val;
      fields[i] = f;
    }
    
    return fields;
  }

  /**
   * Make an Obj of the specified type from the current
   * row in the result set.
   *
   * @param rs The result set to be mapped back into objects
   * @param of The type of the returned object
   * @param fields The mapping of result set columns to fields.
   *   The order and number of fields is the same as the columns
   *   in the result set.
   */
  Obj makeTypedRow(ResultSet rs, Type of, Field[] fields)
    throws SQLException
  {
    if (of.isDynamic().val)
      throw SqlErr.make("Expecting static type, not " + of).val;
    
    Obj row = of.make();
    int numCols = fields.length;
    for (int i=0; i<numCols; ++i)
      fields[i].set(row, SqlUtil.toObj(rs, i+1));
    return row;
  }

//////////////////////////////////////////////////////////////////////////
// Execute
//////////////////////////////////////////////////////////////////////////

  public Int execute(Connection self, Str sql)
  {
    java.sql.Statement stmt = null;
    try
    {
      stmt = jconn.createStatement();
      return Int.make(stmt.executeUpdate(sql.val));
    }
    catch (SQLException e)
    {
      throw err(e);
    }
    finally
    {
      try { if (stmt != null) stmt.close(); } catch (Exception e) {}
    }
  }

//////////////////////////////////////////////////////////////////////////
// Transactions
//////////////////////////////////////////////////////////////////////////

  public Bool getAutoCommit(Connection self)
  {
    try
    {
      return Bool.make(jconn.getAutoCommit());
    }
    catch (SQLException e)
    {
      throw err(e);
    }
  }

  public void setAutoCommit(Connection self, Bool b)
  {
    try
    {
      jconn.setAutoCommit(b.val);
    }
    catch (SQLException e)
    {
      throw err(e);
    }
  }

  public void commit(Connection self)
  {
    try
    {
      jconn.commit();
    }
    catch (SQLException e)
    {
      throw err(e);
    }
  }

  public void rollback(Connection self)
  {
    try
    {
      jconn.rollback();
    }
    catch (SQLException e)
    {
      throw err(e);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Static Init
//////////////////////////////////////////////////////////////////////////

  static
  {
    // preload the driver classes defined in sys.props
    try
    {
      Str drivers = (Str)Sys.env().get(Str.make("jdbc.drivers"), Str.make(""));
      StringTokenizer st = new StringTokenizer(drivers.val, ";");
      while (st.hasMoreTokens())
      {
        String driver = st.nextToken();
        try
        {
          Class.forName(driver);
        }
        catch (Exception e)
        {
          System.out.println("WARNING: Cannot preload jdbc.driver: " + driver);
        }
      }
    }
    catch (Exception e)
    {
      System.out.println(e);
    }
  }

  static Type rowType;
  static List listOfRow;
  static
  {
    try
    {
      rowType = Type.find("sql::Row", true);
      listOfRow = new List(Sys.TypeType, new Type[] { rowType });
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  static RuntimeException err(SQLException e)
  {
    return SqlErr.make(Str.make(e.getMessage()), Err.make(e)).val;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  java.sql.Connection jconn;
}