//
// Copyright (c) 2007, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 07  John Sublett  Creation
//
package fan.sql;

import java.sql.*;
import java.util.HashMap;
import java.util.Iterator;
import fan.sys.*;
import fan.sys.Float;
import fan.sql.Connection;
import fan.sql.Statement;

public class StatementPeer
{
  public static StatementPeer make(Statement fan)
  {
    return new StatementPeer();
  }
  
  Statement prepare(Statement self)
  {
    // Fan uses the ADO .net prepared statement syntax, so
    // for Java the sql needs to be parsed and the @param
    // syntax must be replaced with ?.  It's not a simple
    // replace though because we need to keep the key/value
    // map.
    parse(self.sql.val);
    try
    {
      stmt = self.conn.peer.jconn.prepareStatement(prepared);
    }
    catch(SQLException ex)
    {
      throw ConnectionPeer.err(ex);
    }
    return self;
  }
  
  public List query(Statement self, Map params, Type of)
  {
    try
    {
      setParameters(params);
      ResultSet rs = stmt.executeQuery();
      return self.conn.peer.toRows(rs, of);
    }
    catch(SQLException ex)
    {
      throw ConnectionPeer.err(ex);
    }
  }
  
  public void queryEach(Statement self, Map params, Type of, Func eachFunc)
  {
    try
    {
      setParameters(params);
      ResultSet rs = stmt.executeQuery();
      self.conn.peer.each(rs, of, eachFunc);
    }
    catch(SQLException ex)
    {
      throw ConnectionPeer.err(ex);
    }
  }
  
  public Int execute(Statement self, Map params)
  {
    try
    {
      setParameters(params);
      return Int.make(stmt.executeUpdate());
    }
    catch(SQLException ex)
    {
      throw ConnectionPeer.err(ex);
    }
  }

  /**
   * Set the parameters for the underlying prepared statement
   * using the values specified in the map.
   */
  private void setParameters(Map params)
  {
    try
    {
      Iterator i = paramMap.entrySet().iterator();
      while (i.hasNext())
      {
        java.util.Map.Entry entry = (java.util.Map.Entry)i.next();
        Str key = (Str)entry.getKey();
        Obj value = params.get(key);
        Object jobj = fanToJava(value);
        int[] locs = (int[])entry.getValue();
        for (int j = 0; j < locs.length; j++)
        {
          stmt.setObject(locs[j], jobj);
        }
      }
    }
    catch(SQLException ex)
    {
      throw ConnectionPeer.err(ex);
    }
  }
  
  /**
   * Get a Java object for the specified fan object.
   */
  private Object fanToJava(Obj value)
  {
    Object jobj = value;
    
    // TODO: there's got to be a better way, it'll
    // probably shake out in the ORM design
    if (value instanceof Float)
      jobj = new Double(((Float)value).val);
    else if (value instanceof Int)
      jobj = new Long(((Int)value).val);
    else if (value instanceof Str)
      jobj = ((Str)value).val;
    
    return jobj;
  }
  
  public void close(Statement self)
  {
    try
    {
      stmt.close();
    }
    catch(SQLException ex)
    {
      throw ConnectionPeer.err(ex);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Parse
//////////////////////////////////////////////////////////////////////////

  private void parse(String sql)
  {
    StringBuffer jsql = new StringBuffer(sql.length());
    int index = sql.indexOf('@');
    
    // make sure the sql has at least one parameter
    // before bothering with the parse
    if (index == -1)
    {
      prepared = sql;
      paramMap = new HashMap();
      return;
    }

    Tokenizer t = new Tokenizer(sql);
    String s;
    int pIndex = 1;
    while ((s = t.next()) != null)
    {
      if (s.length() == 0) continue;
      if (s.charAt(0) == '@')
      {
        if (s.length() == 1)
          jsql.append(s);
        else
        {
          if (paramMap == null) paramMap = new HashMap();
          
          // param
          Str key = Str.make(s.substring(1));
          int[] locs = (int[])paramMap.get(key);
          if (locs == null)
          {
            locs = new int[] { pIndex };
            paramMap.put(key, locs);
          }
          else
          {
            int[] temp = new int[locs.length+1];
            System.arraycopy(locs, 0, temp, 0, locs.length);
            temp[locs.length] = pIndex;
            paramMap.put(key, temp);
          }
          pIndex++;
          jsql.append("?");
        }
      }
      else
        jsql.append(s);
    }
    
    prepared = jsql.toString();
  }

//////////////////////////////////////////////////////////////////////////
// Tokenizer
//////////////////////////////////////////////////////////////////////////
  
  private class Tokenizer
  {
    public Tokenizer(String sql)
    {
      this.sql = sql;
      len = sql.length();
      current = 0;
    }
    
    public String next()
    {
      switch (mode)
      {
        case MODE_TEXT: return text();
        case MODE_PARAM: return param();
        case MODE_QUOTE: return quotedText();
        case MODE_END: return null;
        
        default: return null;
      }
    }
    
    private String text()
    {
      int start = current;
      while (current != len)
      {
        int ch = sql.charAt(current);
        if (ch == '@') { mode = MODE_PARAM; break; }
        if (ch == '\'') { mode = MODE_QUOTE; break; }

        current++;
      
        if (current == len) mode = MODE_END;
      }

      return sql.substring(start, current);
    }
    
    private String param()
    {
      int start = current;
      current++;
      
      if (current == len)
        throw SqlErr.make("Invalid parameter.  Parameter name required.").val;
      
      int ch = sql.charAt(current);
      // @@ means we really wanted @
      if (sql.charAt(current) == '@')
      {
        current++;
        return "@";
      }
      
      while (current != len)
      {
        ch = sql.charAt(current);
        boolean valid =
          ((ch >= 'a') && (ch <= 'z')) ||
          ((ch >= 'A') && (ch <= 'Z')) ||
          ((ch >= '0') && (ch <= '9'));
        if (!valid)
        {
          if (ch == '\'')
          {
            mode = MODE_QUOTE;
            break;
          }
          else
          {
            mode = MODE_TEXT;
            break;
          }
        }
        current++;
        if (current == len) mode = MODE_END;
      }
      
      if (current == start+1)
        throw SqlErr.make("Invalid parameter.  Parameter name required.").val;
      
      return sql.substring(start, current);
    }
    
    private String quotedText()
    {
      int start = current;
      int end = -1;
      current++;
      
      if (current == len)
        throw SqlErr.make("Unterminated quoted text.  Expecting '.").val;
      
      while (current != len)
      {
        int ch = sql.charAt(current);
        if (ch == '\'')
        {
          end = current;
          current++;
          break;
        }
        
        current++;
      }
      
      if (end == -1)
        throw SqlErr.make("Unterminated quoted text. Expecting '.").val;
      
      if (current == len)
        mode = MODE_END;
      else
      {
        int ch = sql.charAt(current);
        if (ch == '@')
          mode = MODE_PARAM;
        else if (ch == '\'')
          mode = MODE_QUOTE;
        else
          mode = MODE_TEXT;
      }
      
      return sql.substring(start, end+1);
    }
    
    String sql;
    int    mode = MODE_TEXT;
    int    len;
    int    current;
  }
  
  private static final int MODE_TEXT  = 0;
  private static final int MODE_QUOTE = 1;
  private static final int MODE_PARAM = 2;
  private static final int MODE_END   = 3;
  
  private String            prepared;
  private PreparedStatement stmt;
  private HashMap           paramMap;
}