//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   30 Jun 07  Brian Frank  Creation
//

**
** SqlTest: this test harness assumes MySQL 5.0 running locally.
**
class SqlTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Top
//////////////////////////////////////////////////////////////////////////

  Void test()
  {
    try
    {
      if (!verifyOpen) return
      verify(!c.isClosed())
      dropTables
      createTable
      insertTable
      closures
      transactions
      statements
    }
    finally
    {
      if (c != null)
      {
        c.close
        verify(c.isClosed())
      }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Open
//////////////////////////////////////////////////////////////////////////

  Bool verifyOpen()
  {
    try
    {
      c = Connection.open("jdbc:mysql://localhost:3306/sql_test", "root", "root")
      return true
    }
    catch (Err e)
    {
      echo("**")
      echo("** WARNING: Cannot perform SqlTest without local MySQL database")
      echo("**          $e")
      echo("**")
      return false
    }
  }

//////////////////////////////////////////////////////////////////////////
// Drop Tables
//////////////////////////////////////////////////////////////////////////

  **
  ** Drop all tables in the database
  **
  Void dropTables()
  {
    Str[] tables := c.tables.dup
    while (tables.size != 0)
    {
      Int dropped := 0
      tables.each |Str tableName|
      {
        try
        {
          c.execute("drop table $tableName")
          tables.remove(tableName)
          dropped++
        }
        catch(Err e)
        {
        }
      }
      
      if (dropped == 0)
        throw SqlErr.make("All tables could not be dropped.")
    }
  }

//////////////////////////////////////////////////////////////////////////
// Create Table
//////////////////////////////////////////////////////////////////////////

  Void createTable()
  {
    c.execute("create table farmers(
      farmer_id int auto_increment not null,
      name      varchar(255) not null,
      married   bit,
      pet       varchar(255),
      ss        char(4),
      age       tinyint,
      pigs      smallint,
      cows      int,
      ducks     bigint,
      height    float,
      weight    double,
      primary key (farmer_id))
    ")
    
    row := c.tableRow("farmers")
    rowType := row.type
    
    verify(rowType.field("farmer_id", false) != null, "Missing farmer_id")
    verify(rowType.field("name", false) != null,      "Missing name")
    verify(rowType.field("married", false) != null,   "Missing married")
    verify(rowType.field("pet", false) != null,       "Missing pet")
    verify(rowType.field("ss", false) != null,        "Missing ss")
    verify(rowType.field("age", false) != null,       "Missing age")
    verify(rowType.field("pigs", false) != null,      "Missing pigs")
    verify(rowType.field("cows", false) != null,      "Missing cows")
    verify(rowType.field("ducks", false) != null,     "Missing ducks")
    verify(rowType.field("height", false) != null,    "Missing height")
    verify(rowType.field("weight", false) != null,    "Missing weight")
  }

//////////////////////////////////////////////////////////////////////////
// Insert Table
//////////////////////////////////////////////////////////////////////////

  Void insertTable()
  {
    // insert a couple rows
    data := [
      [1, "Alice",   false, "Pooh",     "abcd", 21,   1,   80,  null, 5.3,  120f],
      [2, "Brian",   true,  "Haley",    "1234", 35,   2,   99,   5,   5.7,  140f],
      [3, "Charlie", null,  "Addi",     null,   null, 3,   44,   7,   null, 6.1],
      [4, "Donny",   true,  null,       "wxyz", 40,  null, null, 8,   null, null],
      [5, "John",    true,  "Berkeley", "5678", 35,  null, null, 8,   null, null],
    ]
    data.each |Obj[] row| { insertFarmer(row[1..-1]) }

    // query
    rows := query("select * from farmers order by farmer_id")
    verifyPlayCols(rows.of)
    verifyEq(data.size, rows.size)
    data.each |Obj[] d, Int i| { verifyRow(rows[i], d) }

    // query with type
    farmers := c.query("select * from farmers order by farmer_id", Farmer.type) as Farmer[]
    verifyEq(farmers is Farmer[], true)
    verifyEq(farmers[0] is Farmer, true)
    f := farmers[0]
    verifyEq(f.farmer_id, 1)
    verifyEq(f.name,     "Alice")
    verifyEq(f.married,   false)
    verifyEq(f.pet,      "Pooh")
    verifyEq(f.ss,       "abcd")
    verifyEq(f.age,      21)
    verifyEq(f.pigs,     1)
    verifyEq(f.cows,     80)
    verifyEq(f.ducks,    null)
    verifyEq(f.height,   5.3)
    verifyEq(f.weight,   120.0)
  }

  Void insertFarmer(Obj[] row)
  {
    s := "insert farmers (name, married, pet, ss, age, pigs, cows, ducks, height, weight) values ("
    s += row.join(", ") |Obj o->Str|
    {
      if (o == null) return "null"
      if (o is Str) return "'$o'"
      return o.toStr
    }
    s += ")"
    verifyEq(execute(s), 1)
  }

  Void verifyPlayCols(Type t)
  {
    verifyEq(t.isDynamic, true)
    verifyEq(t.fields.size, 11)
    verifyCol(t.fields[0],  0,  "farmer_id", Int.type,   "INT")
    verifyCol(t.fields[1],  1,  "name",      Str.type,   "VARCHAR")
    verifyCol(t.fields[2],  2,  "married",   Bool.type,  "BIT")
    verifyCol(t.fields[3],  3,  "pet",       Str.type,   "VARCHAR")
    verifyCol(t.fields[4],  4,  "ss",        Str.type,   "CHAR")
    verifyCol(t.fields[5],  5,  "age",       Int.type,   "TINYINT")
    verifyCol(t.fields[6],  6,  "pigs",      Int.type,   "SMALLINT")
    verifyCol(t.fields[7],  7,  "cows",      Int.type,   "INT")
    verifyCol(t.fields[8],  8,  "ducks",     Int.type,   "BIGINT")
    verifyCol(t.fields[9],  9,  "height",    Float.type, "FLOAT")
    verifyCol(t.fields[10], 10, "weight",    Float.type, "DOUBLE")
  }

//////////////////////////////////////////////////////////////////////////
// Query
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Closures
//////////////////////////////////////////////////////////////////////////

  Void closures()
  {
    ages := Int[,]
    c.query("select age from farmers").each |Obj row| { ages.add(row->age) }
    
    rows := [,]
    c.queryEach("select name, age from farmers", null) |Obj row|
    {
      if (row->age != null)
        row->age = (Int)row->age + 10
      rows.add(row)
    }

    rows.each |Obj row, Int i|
    {
      if (row->age != null) verifyEq(ages[i]+10, row->age)
    }
    
    ages.clear
    rows.clear
    c.query("select age from farmers where age > 30").each |Obj row| { ages.add(row->age) }
    Statement stmt := c.prepare("select age from farmers where age > @age")
    stmt.queryEach(["age":30], null) |Obj row|
    {
      if (row->age != null)
        row->age = (Int)row->age + 10
      rows.add(row)
    }
    
    rows.each |Obj row, Int i|
    {
      if (row->age != null) verifyEq(ages[i]+10, row->age)
    }
  }

//////////////////////////////////////////////////////////////////////////
// Transactions
//////////////////////////////////////////////////////////////////////////

  Void transactions()
  {
    verifyEq(c.autoCommit, true)
    c.autoCommit = false
    verifyEq(c.autoCommit, false)
    c.commit

    rows := query("select name from farmers order by name")
    verifyEq(rows.size, 5)
    verifyEq(rows[0]->name, "Alice")
    verifyEq(rows[1]->name, "Brian")
    verifyEq(rows[2]->name, "Charlie")
    verifyEq(rows[3]->name, "Donny")
    verifyEq(rows[4]->name, "John")

    insertFarmer(["Bad", false, "Bad",  "bad!", 21, 1, 80, null, 5.3, 120f])
    c.rollback
    rows = query("select name from farmers order by name")
    verifyEq(rows.size, 5)
    verifyEq(rows[0]->name, "Alice")
    verifyEq(rows[1]->name, "Brian")
    verifyEq(rows[2]->name, "Charlie")
    verifyEq(rows[3]->name, "Donny")
    verifyEq(rows[4]->name, "John")
  }

//////////////////////////////////////////////////////////////////////////
// Statements
//////////////////////////////////////////////////////////////////////////

  Void statements()
  {
    stmt := c.prepare("select name, age from farmers where name = @name")
    result := stmt.query(["name":"Alice"])
    verifyEq(result[0]->name, "Alice");
    result = stmt.query(["name":"Brian"])
    verifyEq(result[0]->name, "Brian");
    result = stmt.query(["name":"Charlie"])
    verifyEq(result[0]->name, "Charlie");
    result = stmt.query(["name":"Donny"])
    verifyEq(result[0]->name, "Donny");
    result = stmt.query(["name":"John"])
    verifyEq(result[0]->name, "John");
    stmt.close()

    stmt = c.prepare("select name, age from farmers where age > @age")
    result = stmt.query(["age":30])
    verifyEq(result.size, 3)
    result.each |Obj row| { verify(result[0]->age > 30, result[0]->age + " <= 30") }
    stmt.close()
    
    stmt = c.prepare("select name, age from farmers where name = @name and age = @age")
    result = stmt.query(["name":"John", "age":35])
    verifyEq(result.size, 1)
    verifyEq(result[0]->name, "John")
    verifyEq(result[0]->age, 35)
    stmt.close()
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  Row[] query(Str sql)
  {
    rows := (Row[])c.query(sql)
    echo("  q> $sql ($rows.size rows)")
    rows.each |Row row| { echo("     $row") }
    return rows
  }

  Int execute(Str sql)
  {
    echo("  q> $sql")
    return c.execute(sql)
  }

  Void verifyCol(Field f, Int index, Str name, Type of, Str sqlType)
  {
    col := f as Col
    verifyEq(col.index,   index)
    verifyEq(col.name,    name)
    verifySame(col.of,    of)
    verifyEq(col.sqlType, sqlType)
  }

  Void verifyRow(Obj row, Obj[] cells)
  {
    r := row as Row
    verifyEq(r.type.fields.size, cells.size)
    r.type.fields.each |Field f, Int i|
    {
      verifyEq(f.get(r), cells[i])
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Connection c

}

**************************************************************************
** Farmer
**************************************************************************

internal class Farmer
{
  Int farmer_id
  Str name
  Bool married
  Str ss
  Str pet
  Int age
  Num pigs
  Num cows
  Num ducks
  Float weight
  Float height
}

