//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Nov 07  Brian Frank  Creation
//

**
** Resource models an object in the Uri namespace.  Resources
** can be both containers for other resources and a container
** for a Fan object.
**
abstract class Resource
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  **
  ** Constructor for the specified Uri.
  ** Throw ArgErr if uri is null.
  **
  protected new make(Uri uri)

//////////////////////////////////////////////////////////////////////////
// Naming
//////////////////////////////////////////////////////////////////////////

  **
  ** Return if this resource exists in the namespace.
  **
  Bool exists()

  **
  ** Return the Uri of this resource which identifies
  ** its within the local VM namespace.
  **
  Uri uri()

  **
  ** Convenience for [uri.isDir]`Uri.isDir`
  **
  Bool isDir()

  **
  ** Convenience for [uri.path]`Uri.path`.
  **
  Str[] path()

  **
  ** Convenience for [uri.pathStr]`Uri.pathStr`.
  **
  Str pathStr()

  **
  ** Convenience for [uri.name]`Uri.name`.
  **
  Str name()

  **
  ** Convenience for [uri.basename]`Uri.basename`.
  **
  Str basename()

  **
  ** Convenience for [uri.ext]`Uri.ext`.
  **
  Str ext()

//////////////////////////////////////////////////////////////////////////
// Contents
//////////////////////////////////////////////////////////////////////////

  **
  ** Get or set this resource's contents as a Fan object.
  ** This object is always a thread-safe copy of the object.
  ** Use the `save` method to save this object as the
  ** actual resource.
  **
  virtual Obj obj

//////////////////////////////////////////////////////////////////////////
// Children
//////////////////////////////////////////////////////////////////////////

  **
  ** List the children resources contained by this resource.
  ** If this resource doesn't contain children, then return
  ** an empty list.  Note this can potentially be an expensive
  ** operation.  The default implementation returns an empty
  ** list.
  **
  virtual Resource[] list()

  **
  ** Refresh this resource to reflect the current state of
  ** the contents.  This method will typically overwrite the
  ** value of the `obj` field.  The default implementation
  ** throws UnsupportedErr.
  **
  virtual Void refresh()

  **
  ** Create this resource under the namespace.  If there is a
  ** uri conflict, type conflict, or the resource is already
  ** mounted then throw an Err.  Return this.  The default
  ** implementation throws UnsupportedErr.
  **
  virtual Resource create()

  **
  ** Create a child resource under this parent resource
  ** which wraps the specified object.  Return the newly
  ** created resource.
  **
  virtual Resource add(Obj obj)

  **
  ** Save the thread-safe copy of `obj` back to the namespace.
  ** If between the last read and the update, another thread
  ** or application has modified the object then throw an Err.
  ** The default implementation throws UnsupportedErr.
  **
  ** TODO: throw VersionErr, might be a type likely to be caught
  **
  virtual Void save()

  **
  ** Delete this resource from the namespace.  If this resource
  ** contains a child tree of resources, then recursively delete
  ** the entire tree.  The default implementation throws
  ** UnsupportedErr.
  **
  virtual Void delete()

  **
  ** Resolve the Uri to its target as a thread safe working
  ** instance of the resource.  Return null if it can't be
  ** resolved.  The depth indicates the current path depth to
  ** be resolved.  The default implementation returns null.
  **
  virtual Resource resolve(Uri uri, Int startDepth)

}

**************************************************************************
** RootResource
**************************************************************************

internal class RootResource : Resource
{
  internal new make(Uri uri)
}

**************************************************************************
** SysResource
**************************************************************************

internal class SysResource : Resource
{
}
