//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Mar 07  Brian Frank  Creation
//

**
** ResourceTest
**
class ResourceTest : Test
{

  Void testRoot()
  {
    verifyEq(Sys.namespace.uri, `/`)
    verifyEq(Sys.namespace.obj, null)
    verifyEq(`/`.resolve.uri, `/`)
    verify(`/`.resolve.list.any |Resource r->Bool| { return r.uri == `/sys/` })
    verifyErr(UnsupportedErr.type) |,| { `/`.resolve.delete }
    verifyErr(UnsupportedErr.type) |,| { Sys.namespace.delete }

    verifyEq(`/a`.resolve(null, false), null)
    verifyErr(UnresolvedErr.type) |,| { `/a`.resolve }
  }

  Void testAppDir()
  {
    (tempDir + `test.txt`).out.print("Dark Tower").close

    f := Sys.appDir + `test.txt`
    verify(f.exists)
    verifyEq(f.readAllStr, "Dark Tower")
    verify(f.uri.toStr.endsWith("/tmp/test/test.txt"))

    f = (File)`/test.txt`.resolve
    verify(f.exists)
    verifyEq(f.readAllStr, "Dark Tower")
    verifyEq(f.uri, `/test.txt`)
    verify(`/`.resolve.list.any |Resource r->Bool| { return r.uri == `/test.txt` })

    f = tempDir + `dirA/dirB/`
    f.create
    f = (File)`/dirA/dirB/`.resolve
    verify(f.exists)
    verify(f.isDir)
    verifyEq(f.uri, `/dirA/dirB/`)

    f = (File)`/dirA/dirB`.resolve // without trailing slash
    verifyEq(f.uri, `/dirA/dirB/`)

    f = (File)`/dirA/`.resolve.list.first
    verify(f.exists)
    verify(f.isDir)
    verifyEq(f.uri, `/dirA/dirB/`)
  }

  Void testMem()
  {
    // create a, verify unmounted
    a := MemResource.make(`/a`, "-")
    verifyEq(a.exists, false)
    verifyEq(a.uri, `/a`)
    verifyEq(a.name, "a")
    verifyEq(a.path, ["a"])
    verifyEq(a.pathStr, "/a")
    verifyEq(a.list, Resource[,])
    verifyErr(UnsupportedErr.type) |,| { a.refresh }
    verifyErr(UnsupportedErr.type) |,| { a.save }
    verifyErr(UnsupportedErr.type) |,| { a.delete }
    verifyErr(Err.type) |,| { MemResource.make(`/a/x`, "x").create }

    // verify changing obj before mount
    verifyEq(a.obj, "-")
    a.obj = "orig"
    verifyEq(a.obj, "orig")

    // add a
    a.create
    func := |Resource x, Obj obj|
    {
      verifyEq(x.exists, true)
      verifyEq(x.uri, `/a`)
      verifyEq(x.name, "a")
      verifyEq(x.path, ["a"])
      verifyEq(x.pathStr, "/a")
      verify(`/a`.resolve !== x)
      verifyEq(`/a`.resolve.uri, `/a`)
      verifyEq(`/a`.resolve.obj, obj)
    }
    func(a, "orig")
    func(`/a`.resolve, "orig")

    // verify ns.list
    inList := Sys.namespace.list.find |Resource r->Bool| { return r.name == "a" }
    verify(inList != null)
    func(inList, "orig")

    // update via original reference
    a.obj = "updated!"
    func(a, "orig")
    func(`/a`.resolve, "orig")
    a.save
    func(a, "updated!")
    func(`/a`.resolve, "updated!")

    // verify via resolved reference
    r := `/a`.resolve
    verify(r !== a)
    r.obj = "updated again!"
    func(a, "updated!")
    func(`/a`.resolve, "updated!")
    verifySame(`/a`.resolve.obj, "updated!")
    r.save
    func(a, "updated again!")
    func(`/a`.resolve, "updated again!")

    // verify original reference cannot update due to wrong version
    verifyEq(a.obj, "updated!")
    a.obj = "no can do"
    verifyErr(Err.type) |,| { a.save }
    verifyEq(a.uri, `/a`)
    verifyEq(a.obj, "no can do")

    // verify refresh
    a.refresh
    verifyEq(a.uri, `/a`)
    verifyEq(`/a`.resolve.obj, "updated again!")
    func(a, "updated again!")
    a.obj = "change after refresh"
    a.save
    verifyEq(`/a`.resolve.obj, "change after refresh")
    verifyEq(a.obj, "change after refresh")

    // change a to serializable, verify resolve/save
    aval := SerA.make
    a = `/a`.resolve as MemResource
    a.obj = aval
    a.save
    verifySame(a.obj,  aval)
    verifyNotSame(`/a`.resolve.obj, aval)
    verifyEq(`/a`.resolve.obj, aval)
    func(a, aval)
    aval.i = 999
    verifyNotSame(`/a`.resolve.obj, aval)
    verifyEq(`/a`.resolve.obj, SerA.make)
    a.save
    verifyNotSame(`/a`.resolve.obj, aval)
    verifyEq(`/a`.resolve.obj, aval)
    verifyEq(`/a`.resolve.obj, SerA { i = 999 })

    // remove a
    a.delete
    verifyErr(UnresolvedErr.type) |,| { `/a`.resolve }

    // verify our removed resource is reusable
    verifyEq(a.uri, `/a`)
    verifyEq(a.name, "a")
    verifyEq(a.obj, SerA { i = 999 })
    a.create
    func(a, SerA { i = 999 })

    // cleanup
    a.delete
    verifyErr(UnresolvedErr.type) |,| { `/a`.resolve }
  }

  // TODO
  //  - test serializable
  //  - test list
  //  - test delete
  //  - test resolve against base

}