//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   28 Jun 06  Brian Frank  Creation
//

**
** UriTest
**
class UriTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Equals
//////////////////////////////////////////////////////////////////////////

  Void testEquals()
  {
    // equals
    verifyEq(Uri.fromStr("http://u@foo:88/a/b?q#f"), Uri.fromStr("http://u@foo:88/a/b?q#f"));
    verifyEq(`http://foo/`, Uri.fromStr("http://foo/"));
    verifyEq(`http://u@foo:88/a/b?q#f`, `http://u@foo:88/a/b?q#f`);
    verifySame(`http://u@foo:88/a/b?q#f`, `http://u@foo:88/a/b?q#f`);

    // case difference in each component
    verifyEq(`Http://u@foo:88/a/b?q#f`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://u@foo:88/A/b?q#f`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://u@foo:88/a/b?Q#f`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://u@foo:88/a/b?q#F`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://Foo/`, `http://foo/`);
  }

//////////////////////////////////////////////////////////////////////////
// Kitchen Sink
//////////////////////////////////////////////////////////////////////////

  Void testKitchenSink()
  {
    uri := `http://user@host:88/a/b/c.txt?a=b&c=d#frag`
    verifyEq(uri.scheme,   "http")
    verifyEq(uri.auth,     "user@host:88")
    verifyEq(uri.host,     "host")
    verifyEq(uri.port,     88)
    verifyEq(uri.userInfo, "user")
    verifyEq(uri.path,     ["a", "b", "c.txt"])
    verifyEq(uri.pathStr,  "/a/b/c.txt")
    verifyEq(uri.name,     "c.txt")
    verifyEq(uri.basename, "c")
    verifyEq(uri.ext,      "txt")
    verifyEq(uri.query,    ["a":"b", "c":"d"])
    verifyEq(uri.queryStr, "a=b&c=d")
    verifyEq(uri.frag,     "frag")
  }

//////////////////////////////////////////////////////////////////////////
// Encoding
//////////////////////////////////////////////////////////////////////////

  Void testEncoding()
  {
    s  := "a+b://c + \u30A2@d + \u00c0/a:b;/%\u009a%?a=b+ \u00ff#f[?g"
    u1 := `a+b://c + \u30A2@d + \u00c0/a:b;/%\u009a%?a=b+ \u00ff#f[?g`
    u2 := Uri.fromStr(s);
    [u1, u2].each |Uri u|
    {
      verifyEq(s.toStr, s)
      verifyEq(u.scheme, "a+b")
      verifyEq(u.userInfo, "c + \u30A2")
      verifyEq(u.host, "d + \u00c0")
      verifyEq(u.host, "d + \u00c0")
      verifyEq(u.pathStr, "/a:b;/%\u009a%")
      verifyEq(u.path[0], "a:b;")
      verifyEq(u.path[1], "%\u009a%")
      verifyEq(u.queryStr, "a=b+ \u00ff")
      verifyEq(u.query["a"], "b+ \u00ff")
      verifyEq(u.frag, "f[?g")
      verifyEq(u.encode, "a+b://c%20+%20%E3%82%A2@d%20+%20%C3%80/a:b;/%25%C2%9A%25?a=b%2B+%C3%BF#f%5B?g")
      verifyEq(Uri.decode(u.encode), u)
      verifyEq(Uri.decode(u.encode.lower), u)
      verifyEq(Uri.decode(u.encode).toStr, s)
    }

    x := Uri.decode("/foo.png?q_w#f-g~h")
    verifyEq(x.path[0], "foo.png")
    verifyEq(x.basename, "foo")
    verifyEq(x.ext, "png")
    verifyEq(x.queryStr, "q_w")
    verifyEq(x.frag, "f-g~h")

    verifyErr(ParseErr.type) |,| { Uri.decode(s) }
    verifyErr(ParseErr.type) |,| { Uri.decode("/a%") }
    verifyErr(ParseErr.type) |,| { Uri.decode("/a%C") }
    verifyErr(ParseErr.type) |,| { Uri.decode("http://user^host/") }
    verifyErr(ParseErr.type) |,| { Uri.decode("http://foo:aa/") }
    verifyErr(ParseErr.type) |,| { Uri.decode("http://foo/a?h g") }
    verifyErr(ParseErr.type) |,| { Uri.decode("a b") }
    verifyErr(ParseErr.type) |,| { Uri.decode("a#g#h") }
  }

//////////////////////////////////////////////////////////////////////////
// Query Encoding
//////////////////////////////////////////////////////////////////////////

  Void testQueryEncoding()
  {
    verifyQueryEncoding(
      "d=egg&coo&a=b",
      ["d":"egg", "coo":"true", "a":"b"])

    verifyQueryEncoding(
      "a=boo+hoo&%E3%82%A2=%c3%a9!",
      ["a":"boo hoo", "\u30A2":"\u00e9!"])
  }

  Void verifyQueryEncoding(Str encoded, Str:Str q)
  {
    a := Uri.decodeQuery(encoded)
    verifyEq(a, q)
    roundtrip := Uri.decodeQuery(Uri.encodeQuery(q))
    verifyEq(roundtrip, q)
  }

//////////////////////////////////////////////////////////////////////////
// Abs Defaults
//////////////////////////////////////////////////////////////////////////

  Void testAbsDefaults()
  {
    uri := `http://host/a/b/c`
    verifyEq(uri.scheme,   "http")
    verifyEq(uri.auth,     "host")
    verifyEq(uri.host,     "host")
    verifyEq(uri.port,     null)
    verifyEq(uri.userInfo, null)
    verifyEq(uri.path,     ["a", "b", "c"])
    verifyEq(uri.pathStr,  "/a/b/c")
    verifyEq(uri.name,     "c")
    verifyEq(uri.basename, "c")
    verifyEq(uri.ext,      null)
    verifyEq(uri.query,    Str:Str[:])
    verifyEq(uri.queryStr, null)
    verifyEq(uri.frag,     null)
  }

//////////////////////////////////////////////////////////////////////////
// Scheme
//////////////////////////////////////////////////////////////////////////

  Void testScheme()
  {
    verifyScheme(`http://foo/`, "http");
    verifyScheme(`HTTP://foo/`, "http");
    verifyScheme(`mailto:who@there.com`, "mailto");
    verifyScheme(`MailTo:who@there.com`, "mailto");
    verifyScheme(`/a/b/c`, null);
    verifyScheme(`../a/b/c`, null);
    verifyScheme(`#frag`, null);
  }

  Void verifyScheme(Uri uri, Str scheme)
  {
    verifyEq(uri.scheme, scheme)
    verifyEq(uri.isAbs, scheme != null)
    verifyEq(uri.isRel, scheme == null)
  }

//////////////////////////////////////////////////////////////////////////
// Auth (UserInfo, Host, Port)
//////////////////////////////////////////////////////////////////////////

  Void testAuth()
  {
    // combinations
    verifyAuth(`http://brian:pass@host:88/`, "brian:pass@host:88", "brian:pass", "host", 88)
    verifyAuth(`http://brian:pass@host/`, "brian:pass@host", "brian:pass", "host", null)
    verifyAuth(`http://host:88/`, "host:88", null, "host", 88)
    verifyAuth(`http://host/`, "host", null, "host", null)

    // host names
    verifyAuth(`http://HOST/#frag`, "HOST", null, "HOST", null)
    verifyAuth(`foo://www.foo.com:8080/a?q`, "www.foo.com:8080", null, "www.foo.com", 8080)
    verifyAuth(`http://www.foo-bar.com/index`, "www.foo-bar.com", null, "www.foo-bar.com", null)

    // IPv4 addresses
    verifyAuth(`http://10.89.255.0/#frag`, "10.89.255.0", null, "10.89.255.0", null)
    verifyAuth(`http://user@129.5.255.90:90/a/b`, "user@129.5.255.90:90", "user", "129.5.255.90", 90)
    // TODO - Java silently ignores invalid IP addresses like 256.256.256.256
    // verifyAuth(`http://256.256.256.256/`, "256.256.256.256", null, "256.256.256.256", null)

    // IPv6 addresses
    verifyAuth(`http://u@[2001:0db8:0000:0000:0000:0000:1428:57ab]:81/`, "u@[2001:0db8:0000:0000:0000:0000:1428:57ab]:81", "u", "[2001:0db8:0000:0000:0000:0000:1428:57ab]", 81)
    verifyAuth(`ftp://[2001:0db8:0000:0000:0000:0000:1428:57ab]/`, "[2001:0db8:0000:0000:0000:0000:1428:57ab]", null, "[2001:0db8:0000:0000:0000:0000:1428:57ab]", null)
    verifyAuth(`http://[::192.9.5.5]/`, "[::192.9.5.5]", null, "[::192.9.5.5]", null)

    // relative
    verifyAuth(`//u@foo:0/a/b/c`, "u@foo:0", "u", "foo", 0);
    verifyAuth(`//foo/a/b/c`, "foo", null, "foo", null);
    verifyAuth(`/a/b/c`, null, null, null, null);
    verifyAuth(`../a/b/c`, null, null, null, null);
    verifyAuth(`#frag`, null, null, null, null);
  }

  Void verifyAuth(Uri uri, Str auth, Str userInfo, Str host, Int port)
  {
    verifyEq(uri.auth, auth)
    verifyEq(uri.userInfo, userInfo)
    verifyEq(uri.host, host)
    verifyEq(uri.port, port)
  }

//////////////////////////////////////////////////////////////////////////
// Path
//////////////////////////////////////////////////////////////////////////

  Void testPath()
  {
    // none
    verifyPath(`mailto:me@there.com`, "me@there.com", ["me@there.com"])

    // absolute
    verifyPath(`http://host`, "/", Str[,])  // normalized
    verifyPath(`http://host/`, "/", Str[,])
    verifyPath(`http://host/a`, "/a", ["a"])
    verifyPath(`http://host/a/b`, "/a/b", ["a", "b"])
    verifyPath(`http://host/a/b/`, "/a/b/", ["a", "b"])
    verifyPath(`http://host/_a/_b`, "/_a/_b", ["_a", "_b"])

    // auth absolute
    verifyPath(`/`, "/", Str[,])
    verifyPath(`//host/`, "/", Str[,])
    verifyPath(`//host/a`, "/a", ["a"])
    verifyPath(`//host/a/`, "/a/", ["a"])
    verifyPath(`//host/a/b`, "/a/b", ["a", "b"])
    verifyPath(`//host/_a/_b`, "/_a/_b", ["_a", "_b"])

    // path absolute
    verifyPath(`/`, "/", Str[,])
    verifyPath(`/a`, "/a", ["a"])
    verifyPath(`/a/b`, "/a/b", ["a", "b"])

    // path relative
    verifyPath(``, "", Str[,])
    verifyPath(`a`, "a", ["a"])
    verifyPath(`a/b`, "a/b", ["a", "b"])

    // path dot
    verifyPath(`.`, ".", ["."])
    verifyPath(`./a`, "a", ["a"])
    verifyPath(`./a/b`, "a/b", ["a", "b"])

    // path backup
    verifyPath(`..`, "..", [".."])
    verifyPath(`../a`, "../a", ["..", "a"])
    verifyPath(`../a/b`, "../a/b", ["..", "a", "b"])

    // path backup
    verifyPath(`../..`, "../..", ["..", ".."])
    verifyPath(`../../a`, "../../a", ["..", "..", "a"])
    verifyPath(`../../a/b`, "../../a/b", ["..", "..", "a", "b"])
  }

  Void verifyPath(Uri uri, Str pathStr, Str[] path)
  {
    verifyEq(uri.pathStr, pathStr)
    verifyEq(uri.path, path)
    if (path == null)
    {
      verifyEq(uri.isPathAbs, false)
    }
    else
    {
      verifyEq(uri.isPathAbs, pathStr.startsWith("/"))
      verifyEq(uri.path.isRO, true)
      verifyEq(uri.isDir, pathStr.size > 0 && pathStr[-1] == '/')
    }
  }

//////////////////////////////////////////////////////////////////////////
// Name
//////////////////////////////////////////////////////////////////////////

  Void testName()
  {
    // none
    verifyName(`mailto:me@there.com`, "me@there.com", "me@there", "com")

    // absolute roots
    verifyName(`http://host`, null, null, null)
    verifyName(`http://host/`, null, null, null)
    verifyName(`/`, null, null, null)

    // absolute names
    verifyName(`http://host/a`, "a", "a", null)
    verifyName(`http://host/a/`, "a", "a", null)
    verifyName(`http://host/a.`, "a.", "a", "")
    verifyName(`http://host/a.t`, "a.t", "a", "t")

    // relative paths
    verifyName(`alpha`, "alpha", "alpha", null)
    verifyName(`alpha.`, "alpha.", "alpha", "")
    verifyName(`alpha.txt`, "alpha.txt", "alpha", "txt")
    verifyName(`alpha.txt/`, "alpha.txt", "alpha", "txt")

    // dot
    verifyName(`.`, ".", ".", null)
    verifyName(`..`, "..", "..", null)
    verifyName(`../..`, "..", "..", null)
  }

  Void verifyName(Uri uri, Str name, Str basename, Str ext)
  {
    verifyEq(uri.name, name)
    verifyEq(uri.basename, basename)
    verifyEq(uri.ext, ext)
  }

//////////////////////////////////////////////////////////////////////////
// Query
//////////////////////////////////////////////////////////////////////////

  Void testQuery()
  {
    // none
    none := Str:Str[:]
    verifyQuery(`mailto:brian@there.com`, null, none)
    verifyQuery(`mailto:brian?there?com`, "there?com", Str:Str["there?com":"true"])
    verifyQuery(`http://foo/index`, null, none)
    verifyQuery(`http://foo/index#frag`, null, none)
    verifyQuery(`/index`, null, none)

    // various prefixes
    verifyQuery(`http://foo/index?a=b&c=d#frag`, "a=b&c=d", ["a":"b", "c":"d"])
    verifyQuery(`http://foo/index?a=b&c=d`, "a=b&c=d", ["a":"b", "c":"d"])
    verifyQuery(`/index?a=b;c=d`, "a=b;c=d", ["a":"b", "c":"d"])
    verifyQuery(`file.txt?a=b;c=d`, "a=b;c=d", ["a":"b", "c":"d"])
    verifyQuery(`?a=b&c=d`, "a=b&c=d", ["a":"b", "c":"d"])

    // various combinations
    verifyQuery(`?a`, "a", Str:Str["a":"true"])
    verifyQuery(`?a_b!`, "a_b!", Str:Str["a_b!":"true"])
    verifyQuery(`?a=b`, "a=b", ["a":"b"])
    verifyQuery(`?a=beta`, "a=beta", ["a":"beta"])
    verifyQuery(`?alpha=b`, "alpha=b", ["alpha":"b"])
    verifyQuery(`?alpha=b;`, "alpha=b;", ["alpha":"b"])
    verifyQuery(`?alpha=b&`, "alpha=b&", ["alpha":"b"])
    verifyQuery(`?alpha=b;;&;&`, "alpha=b;;&;&", ["alpha":"b"])
    verifyQuery(`?alpha=b&c`, "alpha=b&c", Str:Str["alpha":"b", "c":"true"])
    verifyQuery(`?a=b&;&charlie;`, "a=b&;&charlie;", Str:Str["a":"b", "charlie":"true"])
  }

  Void verifyQuery(Uri uri, Str queryStr, Str:Str query)
  {
    verifyEq(uri.queryStr, queryStr)
    verifyEq(uri.query, query)
    verify(uri.query.isRO())
  }

//////////////////////////////////////////////////////////////////////////
// Frag
//////////////////////////////////////////////////////////////////////////

  Void testFrag()
  {
    // none
    verifyFrag(`mailto:brian@there.com`, null)
    verifyFrag(`mailto:brian?there#com`, "com")
    verifyFrag(`http://foo/index`,  null)

    // combos
    verifyFrag(`http://foo/index?a=b&c=d#frag`, "frag")
    verifyFrag(`http://foo/index#frag`, "frag")
    verifyFrag(`//foo/index#f`, "f")
    verifyFrag(`file.txt#f_g`, "f_g")
    verifyFrag(`#h1`, "h1")
  }

  Void verifyFrag(Uri uri, Str frag)
  {
    verifyEq(uri.frag, frag)
  }

//////////////////////////////////////////////////////////////////////////
// Parent
//////////////////////////////////////////////////////////////////////////

  Void testParent()
  {
    verifyParent(`http://foo:81/a/b/c/?query#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c?query#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c/?query`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c?query`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c/#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c/`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c`, `http://foo:81/a/b/`)

    verifyParent(`http://foo/a/b/c/`, `http://foo/a/b/`)
    verifyParent(`http://foo/a/b/c`, `http://foo/a/b/`)

    verifyParent(`/a/b/c/`, `/a/b/`)
    verifyParent(`/a/b/c`, `/a/b/`)

    verifyParent(`a/b/c/`, `a/b/`)
    verifyParent(`a/b/c`, `a/b/`)

    verifyParent(`../a/b/c/`, `../a/b/`)
    verifyParent(`../a/b/c#frag`, `../a/b/`)

    verifyParent(`/a/`, `/`)
    verifyParent(`/a`,  `/`)

    verifyParent(`/`, null)
    verifyParent(`a.txt`, null)
    verifyParent(`a`, null)
    verifyParent(``, null)
  }

  Void verifyParent(Uri uri, Uri parent)
  {
    verifyEq(uri.parent, parent)
  }

//////////////////////////////////////////////////////////////////////////
// Norm
//////////////////////////////////////////////////////////////////////////

  Void testNorm()
  {
    // abs
    verifyNorm("http://foo:81/path?q#f", `http://foo:81/path?q#f`)
    verifyNorm("http://foo", `http://foo/`)
    verifyNorm("http://foo:80/x", `http://foo/x`)
    verifyNorm("HTTP://foo:80", `http://foo/`)
    verifyNorm("http://foo/.", `http://foo/`)
    verifyNorm("http://foo/..", `http://foo/..`)
    verifyNorm("http://foo/a/..", `http://foo/`)
    verifyNorm("Http://foo:80/a/../..", `http://foo/..`)
    verifyNorm("HTTP://foo/a/b/c/../..", `http://foo/a/`)

    // relative
    verifyNorm("", ``)
    verifyNorm(".", `.`)
    verifyNorm("..", `..`)
    verifyNorm("a/..", ``)
    verifyNorm("a/../..", `..`)
    verifyNorm("a/b/c/../..", `a/`)
    verifyNorm("a/./b/c/../d..", `a/b/d..`)
    verifyNorm("/a/./b/../../d..", `/d..`)
    verifyNorm("a/./b/../c/d../", `a/c/d../`)

    verifyEq(`.`.path, ["."])
    verifyEq(`.`.pathStr, ".")
    verifyEq(`..`.path, [".."])
    verifyEq(`..`.pathStr, "..")
    verifyEq(`../..`.path, ["..", ".."])
    verifyEq(`../..`.pathStr, "../..")
    verifyEq(`x/.`.path, ["x"])
    verifyEq(`x/.`.pathStr, "x/")
  }

  Void verifyNorm(Str unnorm, Uri norm)
  {
    uri := Uri.fromStr(unnorm)
    verifyEq(uri, norm)
  }

//////////////////////////////////////////////////////////////////////////
// Plus
//////////////////////////////////////////////////////////////////////////

  Void testPlus()
  {
    // abs + abs
    verifyPlus(`http://foo:81/path?q#f`, `http://bar:82/p?q2#f2`, `http://bar:82/p?q2#f2`)
    verifyPlus(`http://foo:81/path?q#f`, `mailto:joe@there.com`, `mailto:joe@there.com`)
    verifyPlus(`http://foo/path`, `http://bar/`, `http://bar/`)

    // rel + abs
    verifyPlus(`/a/b/c`, `http://bar:82/p?q2#f2`, `http://bar:82/p?q2#f2`)
    verifyPlus(`/path?q#f`, `mailto:joe@there.com`, `mailto:joe@there.com`)

    // abs + rel /path
    verifyPlus(`http://foo:81/path?q#f`, `/newpath`, `http://foo:81/newpath`)
    verifyPlus(`http://foo:81/path`, `/a/b/c`, `http://foo:81/a/b/c`)
    verifyPlus(`http://foo/`, `/a?query`, `http://foo/a?query`)

    // abs + rel path
    verifyPlus(`http://foo/path?q#f`, `newpath`, `http://foo/newpath`)
    verifyPlus(`http://foo/path/?q#f`, `newpath`, `http://foo/path/newpath`)
    verifyPlus(`http://foo/`, `a/b/c`, `http://foo/a/b/c`)
    verifyPlus(`http://foo/root`, `a/b/c`, `http://foo/a/b/c`)
    verifyPlus(`http://foo/root/`, `a/b/c`, `http://foo/root/a/b/c`)
    verifyPlus(`http://foo/root/`, `../a`, `http://foo/a`)

    // abs + query
    verifyPlus(`http://foo/path?q#f`, `?newquery`, `http://foo/path?newquery`)
    verifyPlus(`http://foo/path/?q#f`, `?newquery`, `http://foo/path/?newquery`)
    verifyPlus(`http://foo/path/`, `?newquery`, `http://foo/path/?newquery`)

    // abs + frag
    verifyPlus(`http://foo/path?q#f`, `#newfrag`, `http://foo/path?q#newfrag`)
    verifyPlus(`http://foo/`, `#newfrag`, `http://foo/#newfrag`)

    // rel + rel
    verifyPlus(`a/b/c`, `d`, `a/b/d`)
    verifyPlus(`a/b/c/`, `d`, `a/b/c/d`)
    verifyPlus(`a/b/c`, `./d`, `a/b/d`)
    verifyPlus(`a/b/c/`, `./d`, `a/b/c/d`)
    verifyPlus(`a/b/c`, `../d`, `a/d`)
    verifyPlus(`a/b/c/`, `../d`, `a/b/d`)
    verifyPlus(`a/b/c`, `../../d`, `d`)
    verifyPlus(`a/b/c/`, `../../d`, `a/d`)
    verifyPlus(`a/b/c`, `../../../d`, `../d`)
    verifyPlus(`a/b/c/`, `../../../d`, `d`)

    // misc
    verifyPlus(`/c:/dev/fan/`, `tmp/test/`, `/c:/dev/fan/tmp/test/`)
  }

  Void testPlusRfc3986()
  {
    // test cases as defined by RFC 3986
    a := `http://a/b/c/d;p?q`

    // 5.4.1. Normal Examples
    verifyPlus(a, `g:h`     , `g:h`)
    verifyPlus(a, `g`       , `http://a/b/c/g`)
    verifyPlus(a, `./g`     , `http://a/b/c/g`)
    verifyPlus(a, `g/`      , `http://a/b/c/g/`)
    verifyPlus(a, `/g`      , `http://a/g`)
    verifyPlus(a, `//g`     , `http://g/`) // derivation b/c normalization
    verifyPlus(a, `?y`      , `http://a/b/c/d;p?y`)
    verifyPlus(a, `g?y`     , `http://a/b/c/g?y`)
    verifyPlus(a, `#s`      , `http://a/b/c/d;p?q#s`)
    verifyPlus(a, `g#s`     , `http://a/b/c/g#s`)
    verifyPlus(a, `g?y#s`   , `http://a/b/c/g?y#s`)
    verifyPlus(a, `;x`      , `http://a/b/c/;x`)
    verifyPlus(a, `g;x`     , `http://a/b/c/g;x`)
    verifyPlus(a, `g;x?y#s` , `http://a/b/c/g;x?y#s`)
    verifyPlus(a, ``        , `http://a/b/c/d;p?q`)
    verifyPlus(a, `.`       , `http://a/b/c/`)
    verifyPlus(a, `./`      , `http://a/b/c/`)
    verifyPlus(a, `..`      , `http://a/b/`)
    verifyPlus(a, `../`     , `http://a/b/`)
    verifyPlus(a, `../g`    , `http://a/b/g`)
    verifyPlus(a, `../..`   , `http://a/`)
    verifyPlus(a, `../../`  , `http://a/`)
    verifyPlus(a, `../../g` , `http://a/g`)

    // 5.4.1. Abnormal Examples
    verifyPlus(a, `../../../g`    , `http://a/g`)
    verifyPlus(a, `../../../../g` , `http://a/g`)
    // we're not going to support this weird case
    // verifyPlus(a, `/./g`          , `http://a/g`)
    // verifyPlus(a, `/../g`         , `http://a/g`)
    verifyPlus(a, `g.`            , `http://a/b/c/g.`)
    verifyPlus(a, `.g`            , `http://a/b/c/.g`)
    verifyPlus(a, `g..`           , `http://a/b/c/g..`)
    verifyPlus(a, `..g`           , `http://a/b/c/..g`)
    verifyPlus(a, `./../g`        , `http://a/b/g`)
    verifyPlus(a, `./g/.`         , `http://a/b/c/g/`)
    verifyPlus(a, `g/./h`         , `http://a/b/c/g/h`)
    verifyPlus(a, `g/../h`        , `http://a/b/c/h`)
    verifyPlus(a, `g;x=1/./y`     , `http://a/b/c/g;x=1/y`)
    verifyPlus(a, `g;x=1/../y`    , `http://a/b/c/y`)
    verifyPlus(a, `g?y/./x`       , `http://a/b/c/g?y/./x`)
    verifyPlus(a, `g?y/../x`      , `http://a/b/c/g?y/../x`)
    verifyPlus(a, `g#s/./x`       , `http://a/b/c/g#s/./x`)
    verifyPlus(a, `g#s/../x`      , `http://a/b/c/g#s/../x`)
  }

  Void verifyPlus(Uri a, Uri b, Uri r)
  {
    x := a + b
    verifyEq(x, r)
    if (x.path != null)  verify(x.path.isRO)
    verify(x.query.isRO)
  }

//////////////////////////////////////////////////////////////////////////
// Minus
//////////////////////////////////////////////////////////////////////////

  Void testMinus()
  {
    verifyMinus(`http://foo/path?q#f`, `http://bar/`, `http://foo/path?q#f`)
    verifyMinus(`http://foo/path`, `http://foo:88/`, `http://foo/path`)
    verifyMinus(`http://foo/path`, `http://user@foo/`, `http://foo/path`)
    verifyMinus(`http://foo/path?q#f`, `/bar`, `http://foo/path?q#f`)
    verifyMinus(`http://foo/path?q#f`, `bar`,  `http://foo/path?q#f`)

    verifyMinus(`http://bar/`, `http://foo/path?q#f`,  `http://bar/`)
    verifyMinus(`/bar`, `http://foo/path?q#f`, `/bar`)
    verifyMinus(`bar`, `http://foo/path?q#f`, `bar`)

    verifyMinus(`http://foo/a/b/c`,  `http://foo/a/b/c`,  ``)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/a/b`,    `c`)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/a/b/`,   `c`)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/a`,      `b/c`)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/`,       `a/b/c`)

    verifyMinus(`/a/b/c`, `/a/b/c`, ``)
    verifyMinus(`/a/b/c`, `/a/b`,   `c`)
    verifyMinus(`/a/b/c`, `/a`,     `b/c`)
    verifyMinus(`/a/b/c`, `/`,      `a/b/c`)
  }

  Void verifyMinus(Uri a, Uri b, Uri r)
  {
    verifyEq(a - b, r)
  }

//////////////////////////////////////////////////////////////////////////
// Tail
//////////////////////////////////////////////////////////////////////////

  Void testTail()
  {
    // various combinations
    verifyEq(`http://host:55/alpha/beta/charlie?query#frag`.tail, `beta/charlie`)
    verifyEq(`pod://sysTest/a/b.txt`.tail, `b.txt`)

    // default 1
    verifyEq(`a/b/c/d`.tail, `b/c/d`)
    verifyEq(`a/b/c`.tail, `b/c`)
    verifyEq(`a/b`.tail, `b`)
    verifyEq(`a`.tail, ``)
    verifyErr(ArgErr.type) |,| { ``.tail }

    // tail(2)
    verifyEq(`a/b/c/d`.tail(2), `c/d`)
    verifyEq(`a/b/c`.tail(2), `c`)
    verifyEq(`a/b`.tail(2), ``)
    verifyErr(ArgErr.type) |,| { `a`.tail(2) }
    verifyErr(ArgErr.type) |,| { ``.tail(2) }

    // tail(3)
    verifyEq(`a/b/c/d`.tail(3), `d`)
    verifyEq(`a/b/c`.tail(3), ``)
    verifyErr(ArgErr.type) |,| { `a/b`.tail(3) }
    verifyErr(ArgErr.type) |,| { `a`.tail(3) }
    verifyErr(ArgErr.type) |,| { ``.tail(3) }
  }

//////////////////////////////////////////////////////////////////////////
// Obj.trapUri
//////////////////////////////////////////////////////////////////////////

  Void testTrapUri()
  {
    verifyEq(``.get(15), 15)
    verifyEq(`toStr`.get(15), "15")
    verifyEq(`toStr/size`.get(15), 2)
    verifyEq(`toStr/size/negate`.get(15), -2)
  }

//////////////////////////////////////////////////////////////////////////
// NameCheck (isName, checkName)
//////////////////////////////////////////////////////////////////////////

  Void testNameCheck()
  {
    verifyNameCheck("", false)
    verifyNameCheck(".", false)
    verifyNameCheck("..", false)
    verifyNameCheck("a b", false)
    verifyNameCheck("\u00ff", false)
    verifyNameCheck("a:b", false)
    verifyNameCheck("x", true)
    verifyNameCheck("7", true)
    verifyNameCheck("-._~", true)
    verifyNameCheck("Hello_77", true)
  }

  Void verifyNameCheck(Str n, Bool ok)
  {
    verifyEq(Uri.isName(n), ok)
    if (ok) Uri.checkName(n)
    else verifyErr(NameErr.type) |,| { Uri.checkName(n) }
  }

//////////////////////////////////////////////////////////////////////////
// Dump
//////////////////////////////////////////////////////////////////////////

  Void dump(Uri uri)
  {
    echo("uri:        $uri")
    echo("  scheme:   $uri.scheme")
    echo("  auth:     $uri.auth")
    echo("  host:     $uri.host")
    echo("  port:     $uri.port")
    echo("  userInfo: $uri.userInfo")
    echo("  path:     $uri.path")
    echo("  pathStr:  $uri.pathStr")
    echo("  name:     $uri.name")
    echo("  basename: $uri.basename")
    echo("  ext:      $uri.ext")
    echo("  query:    $uri.query")
    echo("  queryStr: $uri.queryStr")
    echo("  frag:     $uri.frag")
  }

}