//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   21 Dec 07  Brian Frank  Creation
//

**
** WebService implements the standard pipeline for
** processing  a web request.
**
abstract const class WebService : Thread
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  **
  ** Constructor with thread name.
  **
  new make(Str name) : super(name) {}

//////////////////////////////////////////////////////////////////////////
// Pipeline
//////////////////////////////////////////////////////////////////////////

  **
  ** Run the standard pipeline of steps needed to process
  ** a web request.  This call routes to a series of method
  ** calls which all begin with "step".
  **
  virtual Void pipeline(WebReq req, WebRes res)
  {
    if (!stepInit(req, res)) return
    try
    {
      if (!stepResource(req, res)) return
      if (!stepWeblet(req, res)) return
      if (!stepService(req, res)) return
    }
    catch (Err err)
    {
      err.trace
    }
    finally
    {
      stepCleanup(req, res)
    }
  }

  **
  ** The init step is responsible for initializing thread
  ** local state before any processing is started.  Return
  ** if processing should continue.
  **
  virtual Bool stepInit(WebReq req, WebRes res)
  {
    if (log.isDebug)
    {
      log.debug("==============================")
      log.debug("stepInit [$req.uri]")
    }

    // mount request, response as thread locals
    Thread.locals["web.req"] = req
    Thread.locals["web.res"] = res
    return true
  }

  **
  ** Resolve the request Uri to a Resource in my
  ** namespace.  Return if processing should continue.
  **
  virtual Bool stepResource(WebReq req, WebRes res)
  {
    if (log.isDebug) log.debug("stepResolve [$req.uri]")

    // attempt to resolve
    r := req.uri.resolve(null, false)
    if (log.isDebug) log.debug("  resolved = $r")

    // if resolved successfully we are done
    if (r != null)
    {
      req.resource = r
      return true
    }

    // if not resolved, return 401
    res.statusCode = 404
    res.headers.clear
    res.headers["Content-Type"] = "text/plain"
    res.out.print("Not found: $req.uri\n")
    return false
  }

  **
  ** Resolve the weblet to use for servicing request on
  ** the resource.  Return if processing should continue.
  **
  virtual Bool stepWeblet(WebReq req, WebRes res)
  {
    if (log.isDebug) log.debug("stepWeblet[$req.uri]")

    // TODO: use fixed weblet for now
    Weblet weblet := null
    if (req.resource is File && !req.resource.isDir)
    {
      weblet = FileWeblet.make
    }
    else
    {
      weblet = Type.find("webui::DefaultPage").make
    }

    if (log.isDebug) log.debug("  weblet = $weblet.type")

    // save to thread local
    Thread.locals["web.weblet"] = weblet
    return true
  }

  **
  ** Use the page and widget to service the request.
  **
  virtual Bool stepService(WebReq req, WebRes res)
  {
    Weblet weblet := Thread.locals["web.weblet"]
    weblet.service
    return true
  }

  **
  ** The cleanup step is always run to cleanup
  ** thread local state.
  **
  virtual Void stepCleanup(WebReq req, WebRes res)
  {
    if (log.isDebug) log.debug("stepCleanup [$req.uri]")

    // cleanup thread locals
    Thread.locals.remove("web.req")
    Thread.locals.remove("web.res")
    Thread.locals.remove("web.weblet")
  }

  internal const static Log log := Log.get("web.pipeline")
}
